# :coding: utf-8

import wiz.registry

#: Unique identifier of the plugin.
IDENTIFIER = "installer"


def install_definitions(paths, registry_target, overwrite=False):
    """Install a definition file to a registry.

    *paths* is the path list to all definition files.

    *registry_target* should be a path to a local registry.

    If *overwrite* is True, any existing definitions in the target registry
    will be overwritten.

    Raises :exc:`wiz.exception.IncorrectDefinition` if data in *paths* cannot
    create a valid instance of :class:`wiz.definition.Definition`.

    Raises :exc:`wiz.exception.DefinitionExists` if definition already exists in
    the target registry and *overwrite* is False.

    Raises :exc:`OSError` if the definition can not be exported in a registry
    local *path*.

    """
    definitions = []

    for path in paths:
        _definition = wiz.load_definition(path)
        definitions.append(_definition)

    wiz.registry.install_to_path(
        definitions, registry_target, overwrite=overwrite
    )


def register(config):
    """Register definition installer callback."""
    config.setdefault("callback", {})
    config["callback"]["install"] = install_definitions



# :coding: utf-8

import os
import errno
import io
import unicodedata
import re
import gzip
import pwd
import getpass

import six

import wiz.exception


def get_name():
    """Fetch user full name from password database entry.

    :return: User name of None if name cannot be returned.

    """
    try:
        return pwd.getpwnam(getpass.getuser()).pw_gecos
    except KeyError:
        return


def export(path, content, compressed=False, overwrite=False):
    """Create file from *content* in *path*.

    :param path: Target path to save the file.

    :param content: Content string to write in *path*.

    :param compressed: Indicate whether exported file should be compressed.
        Default is False.

    :param overwrite: Indicate whether any existing path will be overwritten.
        Default is False.

    :raise: :exc:`wiz.exception.FileExists` if overwrite is False and *path*
        already exists.

    """
    # Ensure that "~" is resolved if necessary and that the relative path is
    # always converted into a absolute path.
    path = os.path.abspath(os.path.expanduser(path))

    if os.path.isfile(path) and not overwrite:
        raise wiz.exception.FileExists("{!r} already exists.".format(path))

    ensure_directory(os.path.dirname(path))

    if compressed:
        with gzip.open(path, "wb") as outfile:
            outfile.write(content)

    else:
        with io.open(path, "w", encoding="utf8") as outfile:
            outfile.write(six.text_type(content))


def is_accessible(folder_path):
    """Indicate whether the folder path is accessible.

    :param folder_path: Path to the folder to analyze.

    :return: Boolean value.

    """
    return os.path.isdir(folder_path) and os.access(folder_path, os.R_OK)


def ensure_directory(path):
    """Ensure directory exists at *path*.

    :param path: Path to the folder to create.

    :raise: :exc:`OSError` is path is a file.

    """
    # Explicitly indicate that path should be a directory as default OSError
    # raised by 'os.makedirs' just indicates that the file exists, which is a
    # bit confusing for user.
    if os.path.isfile(path):
        raise OSError("'{}' should be a directory".format(path))

    try:
        os.makedirs(path)
    except OSError as error:
        if error.errno != errno.EEXIST:
            raise

        if not os.path.isdir(path):
            raise


def sanitize_value(value, substitution_character="_", case_sensitive=True):
    """Return *value* suitable for use with filesystem.

    :param value: String value to sanitize.

    :param substitution_character: Symbol to replace awkward characters with.
        Default is underscore.

    :param case_sensitive: Indicate whether case should be preserved or if
        value should be converted to lowercase. Default is True.

    :return: Sanitized value.

    """
    value = unicodedata.normalize("NFKD", six.u(value))
    value = value.encode("ascii", "ignore").decode("ascii")
    value = re.sub(r"[^\w._\-\\/:%]", substitution_character, value)
    value = value.strip()

    if not case_sensitive:
        value = value.lower()

    return six.text_type(value)



# :coding: utf-8

from __future__ import print_function

import os
import sys
import select
import subprocess
import tempfile
import termios
import tty
import pty
import signal

import wiz.logging
import wiz.utility
import wiz.symbol


def shell(environment, command=None):
    """Spawn a sub-shell with an *environment* mapping.

    :param environment: Environment mapping to spawn the shell with.

    :param command: Mapping of command aliases which should be available in the
        shell.

    .. note::

        Default shell is :term:`Bash`.

    """
    logger = wiz.logging.Logger(__name__ + ".shell")

    if command is None:
        command = {}

    # TODO: Improve default shell to make it cross platform...
    executable = "/bin/bash"
    logger.info("Spawn shell: {}".format(executable))

    # save original tty setting then set it to raw mode
    old_tty = termios.tcgetattr(sys.stdin)
    tty.setraw(sys.stdin.fileno())

    # open pseudo-terminal to interact with subprocess
    master_fd, slave_fd = pty.openpty()

    # Create temporary rc file for shell aliases for commands
    rcfile = tempfile.NamedTemporaryFile()
    for alias, value in command.items():
        line = "alias {0}='{1}'\n".format(alias, value)
        rcfile.write(line.encode("utf-8"))
        rcfile.seek(0)
        rcfile.read()

    if os.path.exists(rcfile.name):
        executable = [executable, "--rcfile", rcfile.name]

    # Run in a new process group to enable job control
    process = subprocess.Popen(
        executable,
        preexec_fn=os.setsid,
        stdin=slave_fd,
        stdout=slave_fd,
        stderr=slave_fd,
        universal_newlines=True,
        env=environment
    )

    # Register the cleanup function as handler for SIGINT and SIGTERM.
    signal.signal(signal.SIGINT, _cleanup)
    signal.signal(signal.SIGTERM, _cleanup)

    while process.poll() is None:
        read_list, write_list, _ = select.select([sys.stdin, master_fd], [], [])

        if sys.stdin in read_list:
            message = os.read(sys.stdin.fileno(), 10240)
            os.write(master_fd, message)

        elif master_fd in read_list:
            message = os.read(master_fd, 10240)
            if message:
                os.write(sys.stdout.fileno(), message)

    # Restore tty settings back
    termios.tcsetattr(sys.stdin, termios.TCSADRAIN, old_tty)

    # Remove temporary rc file for shell aliases
    rcfile.close()


def execute(elements, environment):
    """Run command *elements* within a specific *environment*.

    :param elements: List of strings constituting the command line to execute
        (e.g. ["app_exe", "--option", "value"])

    :param environment: Environment mapping to execute command with.

    """
    logger = wiz.logging.Logger(__name__ + ".shell")
    logger.info(
        "Start command: {}".format(wiz.utility.combine_command(elements))
    )

    # Substitute environment variables from command line elements.
    elements = [
        wiz.environ.substitute(element, environment) for element in elements
    ]

    # Register the cleanup function as handler for SIGINT and SIGTERM.
    signal.signal(signal.SIGINT, _cleanup)
    signal.signal(signal.SIGTERM, _cleanup)

    try:
        subprocess.call(elements, env=environment)
    except OSError:
        logger.error(
            "Executable can not be found within resolved "
            "environment [{}]".format(elements[0])
        )
        logger.debug_traceback()


def _cleanup(signum, frame):
    """Exit from python if a process is terminated or interrupted."""
    sys.exit(0)



# :coding: utf-8

"""
Loading a definition should take around few milliseconds maximum, as the tool
should be able to load >= 5000 definitions in less than a second.

"""

import os

import ujson
import pytest

import wiz
import wiz.config
import wiz.definition


@pytest.fixture(autouse=True)
def reset_configuration(mocker):
    """Ensure that no personal configuration is fetched during tests."""
    mocker.patch.object(os.path, "expanduser", return_value="__HOME__")

    # Reset configuration.
    wiz.config.fetch(refresh=True)


def test_load_minimal(temporary_directory, benchmark):
    """Load a minimal definition."""
    data = {"identifier": "foo"}

    path = os.path.join(temporary_directory, "definition.json")
    with open(path, "w") as stream:
        ujson.dump(data, stream)

    benchmark(wiz.definition.load, path)


def test_load_simple(temporary_directory, benchmark):
    """Load a simple definition."""
    data = {
        "identifier": "foo",
        "version": "0.1.0",
        "description": "This is a definition.",
        "command": {
            "foo": "FooExe"
        },
        "environ": {
            "Key1": "Value1",
            "Key2": "Value2",
            "Key3": "Value3",
        },
        "requirements": [
            "fee >= 0.1.0, < 1",
            "bar >= 2.3, < 3",
            "bim != 6.0.0",
        ]
    }

    path = os.path.join(temporary_directory, "definition.json")
    with open(path, "w") as stream:
        ujson.dump(data, stream)

    benchmark(wiz.definition.load, path)


def test_load_simple_with_variants(temporary_directory, benchmark):
    """Load a simple definition with variants."""
    data = {
        "identifier": "foo",
        "version": "0.1.0",
        "description": "This is a definition.",
        "command": {
            "foo": "FooExe"
        },
        "environ": {
            "Key1": "Value1",
            "Key2": "Value2",
            "Key3": "Value3",
        },
        "requirements": [
            "fee >= 0.1.0, < 1",
            "bar >= 2.3, < 3",
            "bim != 6.0.0",
        ],
        "variants": [
            {
                "identifier": "V4",
                "requirements": [
                    "bew >= 4, < 5",
                ]
            },
            {
                "identifier": "V3",
                "requirements": [
                    "bew >= 3, < 4",
                ]
            },
            {
                "identifier": "V2",
                "requirements": [
                    "bew >= 2, < 3",
                ],
            },
            {
                "identifier": "V1",
                "requirements": [
                    "bew >= 1, < 2",
                ]
            }
        ]
    }

    path = os.path.join(temporary_directory, "definition.json")
    with open(path, "w") as stream:
        ujson.dump(data, stream)

    benchmark(wiz.definition.load, path)


def test_load_many_environments(temporary_directory, benchmark):
    """Load a definition with 1000 environment variables."""
    data = {
        "identifier": "foo",
        "version": "0.1.0",
        "description": "This is a definition.",
        "command": {
            "foo": "FooExe"
        },
        "environ": {
            "KEY{}".format(index): "VALUE{}".format(index)
            for index in range(1000)
        },
        "requirements": [
            "fee >= 0.1.0, < 1",
            "bar >= 2.3, < 3",
            "bim != 6.0.0",
        ]
    }

    path = os.path.join(temporary_directory, "definition.json")
    with open(path, "w") as stream:
        ujson.dump(data, stream)

    benchmark(wiz.definition.load, path)


def test_load_many_requirements(temporary_directory, benchmark):
    """Load a definition with 100 requirements."""
    data = {
        "identifier": "foo",
        "version": "0.1.0",
        "description": "This is a definition.",
        "command": {
            "foo": "FooExe"
        },
        "environ": {
            "Key1": "Value1",
            "Key2": "Value2",
            "Key3": "Value3",
        },
        "requirements": [
            "fee{}".format(index) for index in range(100)
        ]
    }

    path = os.path.join(temporary_directory, "definition.json")
    with open(path, "w") as stream:
        ujson.dump(data, stream)

    benchmark(wiz.definition.load, path)


def test_load_many_variants(temporary_directory, benchmark):
    """Load a definition with 100 variants."""
    data = {
        "identifier": "foo",
        "version": "0.1.0",
        "description": "This is a definition.",
        "command": {
            "foo": "FooExe"
        },
        "environ": {
            "Key1": "Value1",
            "Key2": "Value2",
            "Key3": "Value3",
        },
        "requirements": [
            "fee >= 0.1.0, < 1",
            "bar >= 2.3, < 3",
            "bim != 6.0.0",
        ],
        "variants": [
            {"identifier": "V{}".format(index)}
            for index in range(100)
        ]
    }

    path = os.path.join(temporary_directory, "definition.json")
    with open(path, "w") as stream:
        ujson.dump(data, stream)

    benchmark(wiz.definition.load, path)


def test_load_many_complex_variants(temporary_directory, benchmark):
    """Load a definition with 100 complex variants."""
    data = {
        "identifier": "foo",
        "version": "0.1.0",
        "description": "This is a definition.",
        "command": {
            "foo": "FooExe"
        },
        "environ": {
            "Key1": "Value1",
            "Key2": "Value2",
            "Key3": "Value3",
        },
        "requirements": [
            "fee >= 0.1.0, < 1",
            "bar >= 2.3, < 3",
            "bim != 6.0.0",
        ],
        "variants": [
            {
                "identifier": "V{}".format(index),
                "requirements": ["bew{}".format(index) for index in range(100)],
                "environ": {
                    "VAR_KEY{}".format(index): "VALUE{}".format(index)
                    for index in range(100)
                }
            }
            for index in range(100)
        ]
    }

    path = os.path.join(temporary_directory, "definition.json")
    with open(path, "w") as stream:
        ujson.dump(data, stream)

    benchmark(wiz.definition.load, path)


def test_load_complex(temporary_directory, benchmark):
    """Load a complex definition."""
    data = {
        "identifier": "foo",
        "version": "0.1.0",
        "description": "This is a definition.",
        "command": {
            "foo": "FooExe"
        },
        "environ": {
            "KEY{}".format(index): "VALUE{}".format(index)
            for index in range(1000)
        },
        "requirements": [
            "fee{}".format(index) for index in range(100)
        ],
        "variants": [
            {
                "identifier": "V{}".format(index),
                "requirements": ["bew{}".format(index) for index in range(100)],
                "environ": {
                    "VAR_KEY{}".format(index): "VALUE{}".format(index)
                    for index in range(100)
                }
            }
            for index in range(100)
        ]
    }

    path = os.path.join(temporary_directory, "definition.json")
    with open(path, "w") as stream:
        ujson.dump(data, stream)

    benchmark(wiz.definition.load, path)



# :coding: utf-8

import os
import uuid
import shutil
import tempfile

import ujson
import pytest

import wiz
import wiz.config


@pytest.fixture(autouse=True)
def reset_configuration(mocker):
    """Ensure that no personal configuration is fetched during tests."""
    mocker.patch.object(os.path, "expanduser", return_value="__HOME__")

    # Reset configuration.
    wiz.config.fetch(refresh=True)


@pytest.fixture(scope="module")
def registries(request):
    """Return mocked registry paths with many definitions at several depths."""
    # Create registries
    registry_01 = tempfile.mkdtemp()
    registry_02 = tempfile.mkdtemp()
    registry_03 = tempfile.mkdtemp()

    # Define lots of environment variables for each definitions.
    environ = {
        "KEY{}".format(index): "VALUE{}".format(index)
        for index in range(100)
    }

    # Define lots of commands for each definitions.
    command = {
        "app{}".format(index): "App{}".format(index)
        for index in range(100)
    }

    # Define lots of requirements for each definitions.
    requirements = ["foo{}".format(index) for index in range(100)]

    # Define lots of variants for each definitions.
    variants = [
        {
            "identifier": "Variant{}".format(index),
            "requirements": ["bar{}".format(index) for index in range(100)],
            "environ": {
                "VAR_KEY{}".format(index): "VALUE{}".format(index)
                for index in range(100)
            }
        }
        for index in range(500)
    ]

    # All possible versions and system requirements.
    versions = ["0.1.0", "0.2.0", "1.0.0", "2.0.0", "2.0.0a", None]
    systems = [{"platform": "linux"}, {"platform": "windows"}]

    # Initialize identifier
    identifier = None

    for registry in [registry_01, registry_02, registry_03]:
        for index in range(1500):
            version = versions[index % len(versions)]
            system = systems[index % len(systems)]

            # Reset identifier each time version is None or each time version
            # cycle is re-initiated.
            if (
                identifier is None or version is None
                or not index % len(versions)
            ):
                identifier = str(uuid.uuid4())

            data = {
                "identifier": identifier,
                "system": system,
                "environ": environ,
                "command": command,
                "requirements": requirements,
                "variants": variants[:]
            }

            if version is not None:
                data["version"] = version

            # Write definition
            path = os.path.join(registry, identifier)
            file_path = os.path.join(path, "{}.json".format(identifier))

            if not os.path.isdir(path):
                os.makedirs(path)

            with open(file_path, "w") as stream:
                ujson.dumps(data, stream)

    def cleanup():
        """Remove temporary directories."""
        shutil.rmtree(registry_01)
        shutil.rmtree(registry_02)
        shutil.rmtree(registry_03)

    request.addfinalizer(cleanup)
    return [registry_01, registry_02, registry_03]


def test_discover_1500_definitions(registries, benchmark):
    """Test performance when fetching 1500 definitions in one registry."""
    benchmark(wiz.fetch_definition_mapping, registries[:1])


def test_discover_3000_definitions(registries, benchmark):
    """Test performance when fetching 3000 definitions in two registries."""
    benchmark(wiz.fetch_definition_mapping, registries[:2])


def test_discover_4500_definitions(registries, benchmark):
    """Test performance when fetching 4500 definitions in three registries."""
    benchmark(wiz.fetch_definition_mapping, registries)


def test_discover_4500_definitions_linux_only(registries, benchmark):
    """Test performance when fetching definitions for linux only."""
    benchmark(
        wiz.fetch_definition_mapping, registries,
        system_mapping={"platform": "linux"}
    )


def test_discover_4500_definitions_windows_only(registries, benchmark):
    """Test performance when fetching definitions for windows only."""
    benchmark(
        wiz.fetch_definition_mapping, registries,
        system_mapping={"platform": "windows"}
    )



# :coding: utf-8

import datetime
import os
import tempfile

import click
import pytest
from click.testing import CliRunner
from six.moves import reload_module

import wiz.command_line
import wiz.config
import wiz.definition
import wiz.exception
import wiz.filesystem
import wiz.history
import wiz.package
import wiz.registry
import wiz.spawn
import wiz.symbol
import wiz.utility


@pytest.fixture(autouse=True)
def reset_configuration(mocker):
    """Ensure that no personal configuration is fetched during tests."""
    mocker.patch.object(os.path, "expanduser", return_value="__HOME__")

    # Reset configuration.
    wiz.command_line._CONFIG = wiz.config.fetch(refresh=True)
    reload_module(wiz.command_line)


@pytest.fixture()
def mock_datetime_now(mocker):
    """Return mocked 'wiz.fetch_definition_mapping' function."""
    _date = mocker.Mock(**{"isoformat.return_value": "NOW"})
    return mocker.patch.object(
        datetime, "datetime", **{"now.return_value": _date}
    )


@pytest.fixture()
def mocked_click_edit(mocker):
    """Return mocked 'click.edit' function."""
    return mocker.patch.object(click, "edit")


@pytest.fixture()
def mocked_click_prompt(mocker):
    """Return mocked 'click.prompt' function."""
    return mocker.patch.object(click, "prompt")


@pytest.fixture()
def mocked_click_confirm(mocker):
    """Return mocked 'click.confirm' function."""
    return mocker.patch.object(click, "confirm")


@pytest.fixture()
def mocked_fetch_definition_mapping(mocker):
    """Return mocked 'wiz.fetch_definition_mapping' function."""
    return mocker.patch.object(wiz, "fetch_definition_mapping")


@pytest.fixture()
def mocked_load_definition(mocker):
    """Return mocked 'wiz.load_definition' function."""
    return mocker.patch.object(wiz, "load_definition")


@pytest.fixture()
def mocked_resolve_context(mocker):
    """Return mocked 'wiz.resolve_context' function."""
    return mocker.patch.object(wiz, "resolve_context")


@pytest.fixture()
def mocked_resolve_command(mocker):
    """Return mocked 'wiz.resolve_command' function."""
    return mocker.patch.object(wiz, "resolve_command")


@pytest.fixture()
def mocked_fetch_package_request_from_command(mocker):
    """Return mocked 'wiz.fetch_package_request_from_command' function."""
    return mocker.patch.object(wiz, "fetch_package_request_from_command")


@pytest.fixture()
def mocked_export_definition(mocker):
    """Return mocked 'wiz.export_definition' function."""
    return mocker.patch.object(wiz, "export_definition")


@pytest.fixture()
def mocked_export_script(mocker):
    """Return mocked 'wiz.export_script' function."""
    return mocker.patch.object(wiz, "export_script")


@pytest.fixture()
def mocked_definition_discover(mocker):
    """Return mocked 'wiz.definition.discover' function."""
    return mocker.patch.object(wiz.definition, "discover")


@pytest.fixture()
def mocked_history_start_recording(mocker):
    """Return mocked 'wiz.history.start_recording' function."""
    return mocker.patch.object(wiz.history, "start_recording")


@pytest.fixture()
def mocked_history_record_action(mocker):
    """Return mocked 'wiz.history.record_action' function."""
    return mocker.patch.object(wiz.history, "record_action")


@pytest.fixture()
def mocked_history_get(mocker):
    """Return mocked 'wiz.history.get' function."""
    return mocker.patch.object(wiz.history, "get")


@pytest.fixture()
def mocked_spawn_execute(mocker):
    """Return mocked 'wiz.spawn.execute' function."""
    return mocker.patch.object(wiz.spawn, "execute")


@pytest.fixture()
def mocked_spawn_shell(mocker):
    """Return mocked 'wiz.spawn.shell' function."""
    return mocker.patch.object(wiz.spawn, "shell")


@pytest.fixture()
def mocked_filesystem_export(mocker):
    """Return mocked 'wiz.filesystem.export' function."""
    return mocker.patch.object(wiz.filesystem, "export")


@pytest.fixture()
def mocked_system_query(mocker):
    """Return mocked 'wiz.system.query' function."""
    return mocker.patch.object(wiz.system, "query")


@pytest.fixture()
def mocked_registry_fetch(mocker):
    """Return mocked 'wiz.registry.fetch' function."""
    return mocker.patch.object(wiz.registry, "fetch")


@pytest.fixture()
def mocked_registry_install_to_path(mocker):
    """Return mocked 'wiz.registry.install_to_path' function."""
    return mocker.patch.object(wiz.registry, "install_to_path")


@pytest.fixture()
def definitions():
    """Return mocked definitions."""
    return [
        wiz.definition.Definition(
            {
                "identifier": "foo",
                "version": "0.2.0",
                "description": "This is Foo 0.2.0.",
                "command": {
                    "fooExe": "fooExe -X"
                },
                "system": {
                    "platform": "linux",
                }
            },
            registry_path="/registry1"
        ),
        wiz.definition.Definition(
            {
                "identifier": "foo",
                "version": "0.2.0",
                "description": "This is Foo 0.2.0.",
                "command": {
                    "fooExe": "fooExe -X"
                },
                "system": {
                    "platform": "mac",
                }
            },
            registry_path="/registry1"
        ),
        wiz.definition.Definition(
            {
                "identifier": "foo",
                "version": "0.1.0",
                "description": "This is Foo 0.1.0.",
                "command": {
                    "fooExe": "fooExe -X"
                },
                "system": {
                    "platform": "linux",
                },
                "environ": {
                    "PATH": "/path/to/bin:${PATH}",
                    "PYTHONPATH": "/path/to/lib:${PYTHONPATH}",
                },
                "requirements": [
                    "bim >= 0.1.0, < 1"
                ]
            },
            registry_path="/registry1"
        ),
        wiz.definition.Definition(
            {
                "identifier": "bar",
                "version": "0.1.0",
                "description": "This is Bar 0.1.0.",
                "environ": {
                    "PATH": "/path/to/bin:${PATH}",
                },
                "variants": [
                    {
                        "identifier": "Variant1",
                        "environ": {
                            "PYTHONPATH": "/path/to/lib/1:${PYTHONPATH}",
                        }
                    },
                    {
                        "identifier": "Variant2",
                        "environ": {
                            "PYTHONPATH": "/path/to/lib/2:${PYTHONPATH}",
                        }
                    },
                ],
                "requirements": [
                    "bim >= 0.1.0, < 1"
                ]
            },
            registry_path="/registry2"
        ),
        wiz.definition.Definition(
            {
                "identifier": "bim",
                "version": "0.1.1",
                "description": "This is Bim 0.1.1.",
                "command": {
                    "bimExe": "bimExe -X"
                },
                "variants": [
                    {
                        "identifier": "Variant1",
                        "environ": {
                            "PYTHONPATH": "/path/to/lib/1:${PYTHONPATH}",
                        }
                    },
                ],
            },
            registry_path="/registry2"
        ),
        wiz.definition.Definition(
            {
                "identifier": "bim",
                "version": "0.1.0",
                "description": "This is Bim 0.1.0.",
                "command": {
                    "bimExe": "bimExe -X"
                },
            },
            registry_path="/registry2"
        )
    ]


@pytest.fixture()
def definition_mapping():
    """Return mocked definition mapping."""
    return {
        "command": {
            "fooExe": "foo",
            "bimExe": "bim",
        },
        "package": {
            "foo": {
                "0.2.0": wiz.definition.Definition(
                    {
                        "identifier": "foo",
                        "version": "0.2.0",
                        "description": "This is Foo 0.2.0.",
                    },
                    path="/registry1/foo-0.2.0.json",
                    registry_path="/registry1",
                ),
                "0.1.0": wiz.definition.Definition(
                    {
                        "identifier": "foo",
                        "version": "0.1.0",
                        "description": "This is Foo 0.1.0.",
                        "environ": {
                            "PATH": "/path/to/bin:${PATH}",
                            "PYTHONPATH": "/path/to/lib:${PYTHONPATH}",
                        },
                        "requirements": [
                            "bim >= 0.1.0, < 1"
                        ]
                    },
                    path="/registry1/foo-0.1.0.json",
                    registry_path="/registry1",
                ),
            },
            "bar": {
                "0.1.0": wiz.definition.Definition(
                    {
                        "identifier": "bar",
                        "version": "0.1.0",
                        "description": "This is Bar 0.1.0.",
                        "environ": {
                            "PATH": "/path/to/bin:${PATH}",
                        },
                        "variants": [
                            {
                                "identifier": "Variant1",
                                "environ": {
                                    "PYTHONPATH": (
                                        "/path/to/lib/1:${PYTHONPATH}"
                                    ),
                                }
                            },
                            {
                                "identifier": "Variant2",
                                "environ": {
                                    "PYTHONPATH": (
                                        "/path/to/lib/2:${PYTHONPATH}"
                                    ),
                                }
                            },
                        ],
                        "requirements": [
                            "bim >= 0.1.0, < 1"
                        ]
                    },
                    path="/registry1/bar-0.1.0.json",
                    registry_path="/registry2",
                ),
            },
            "bim": {
                "0.1.1": wiz.definition.Definition(
                    {
                        "identifier": "bim",
                        "version": "0.1.1",
                        "description": "This is Bim 0.1.1.",
                        "variants": [
                            {
                                "identifier": "Variant1",
                                "environ": {
                                    "PYTHONPATH": (
                                        "/path/to/lib/1:${PYTHONPATH}"
                                    ),
                                }
                            },
                        ],
                    },
                    path="/registry1/bim-0.1.1.json",
                    registry_path="/registry2",
                ),
                "0.1.0": wiz.definition.Definition(
                    {
                        "identifier": "bim",
                        "version": "0.1.0",
                        "description": "This is Bim 0.1.0.",
                    },
                    path="/registry1/bim-0.1.0.json",
                    registry_path="/registry2",
                ),
            }
        },
        "registries": ["/registry1", "/registry2"]
    }


@pytest.fixture()
def wiz_context():
    """Return mocked context."""
    foo_definition = wiz.definition.Definition(
        {
            "identifier": "foo",
            "version": "0.1.0",
            "description": "This is Foo 0.1.0.",
            "environ": {
                "PATH": "/path/to/foo/bin:${PATH}",
                "PYTHONPATH": "/path/to/foo/lib:${PYTHONPATH}",
            },
            "requirements": [
                "bim >= 0.1.0, < 1"
            ]
        },
        registry_path="/registry1"
    )

    bim_definition = wiz.definition.Definition(
        {
            "identifier": "bim",
            "version": "0.1.1",
            "description": "This is Bim 0.1.1.",
            "environ": {
                "PATH": "/path/to/bim/bin:${PATH}",
                "PYTHONPATH": "/path/to/bim/lib:${PYTHONPATH}",
                "LICENSE_ENV": "license@bim.com:2000"
            },
            "variants": [
                {
                    "identifier": "Variant1",
                    "environ": {
                        "PYTHONPATH": "/path/to/bim/lib:${PYTHONPATH}",
                    }
                }
            ]
        },
        registry_path="/registry2"
    )

    return {
        "command": {
            "fooExe": "foo",
            "fooExeDebug": "foo --debug",
        },
        "environ": {
            "KEY1": "value1",
            "KEY2": "value2",
        },
        "packages": [
            wiz.package.create(foo_definition),
            wiz.package.create(
                bim_definition, variant_identifier="Variant1"
            ),
        ],
        "registries": ["/registry1", "/registry2"]
    }


def test_empty_arguments(
    mocked_history_start_recording, mocked_history_get,
    mocked_filesystem_export, mocked_system_query,
    mocked_registry_fetch, mocked_fetch_definition_mapping
):
    """Do not raise error for empty arguments."""
    runner = CliRunner()
    result = runner.invoke(wiz.command_line.main)
    assert result.exit_code == 0
    assert not result.exception

    mocked_history_start_recording.assert_not_called()
    mocked_history_get.assert_not_called()
    mocked_filesystem_export.assert_not_called()
    mocked_system_query.assert_not_called()
    mocked_registry_fetch.assert_not_called()
    mocked_fetch_definition_mapping.assert_not_called()


@pytest.mark.parametrize(
    "options, platform, architecture, os_name, os_version", [
        ([], None, None, None, None),
        (["--platform", "linux"], "linux", None, None, None),
        (["--architecture", "x86_64"], None, "x86_64", None, None),
        (["--os-name", "centos"], None, None, "centos", None),
        (["--os-version", "7.4.1708"], None, None, None, "7.4.1708"),
    ], ids=[
        "no-options",
        "override-platform",
        "override-architecture",
        "override-os-name",
        "override-os-version",
    ]
)
@pytest.mark.usefixtures("mocked_registry_fetch")
@pytest.mark.usefixtures("mocked_fetch_definition_mapping")
def test_query_system(
    mocked_system_query, options, platform, architecture, os_name, os_version
):
    """Override system with options."""
    runner = CliRunner()
    result = runner.invoke(wiz.command_line.main, options + ["list", "package"])
    assert result.exit_code == 0
    assert not result.exception

    mocked_system_query.assert_called_once_with(
        platform=platform,
        architecture=architecture,
        os_name=os_name,
        os_version=os_version,
    )


@pytest.mark.parametrize(
    "options, paths, depth, include_local, include_cwd", [
        ([], None, None, True, True),
        (
            ["-r", "/path1", "-r", "/path2"],
            ["/path1", "/path2"], None, True, True
        ),
        (["-rd", "2"], None, 2, True, True),
        (
            ["-add", "/path2", "-r", "/path1"],
            ["/path1", "/path2"], None, True, True
        ),
        (["--no-local"], None, None, False, True),
        (["--no-cwd"], None, None, True, False),
    ], ids=[
        "no-options",
        "change-search-paths",
        "change-search-depth",
        "add-search-path",
        "skip-local",
        "skip-cwd",
    ]
)
def test_fetch_registry(
    mocked_system_query, mocked_registry_fetch, mocked_definition_discover,
    options, paths, depth, include_local, include_cwd
):
    """Override registries with options."""
    mocked_system_query.return_value = "__SYSTEM__"
    mocked_registry_fetch.return_value = ["/registry1", "/registry2"]

    runner = CliRunner()
    result = runner.invoke(wiz.command_line.main, options + ["list", "package"])
    assert result.exit_code == 0
    assert not result.exception

    mocked_registry_fetch.assert_called_once_with(
        tuple(paths or wiz.registry.get_defaults()),
        include_local=include_local,
        include_working_directory=include_cwd
    )

    mocked_definition_discover.assert_called_once_with(
        ["/registry1", "/registry2"],
        system_mapping="__SYSTEM__",
        max_depth=depth
    )


@pytest.mark.parametrize("options, recorded", [
    ([], False),
    (["--record", tempfile.gettempdir()], True)
], ids=[
    "normal",
    "recorded",
])
@pytest.mark.usefixtures("mock_datetime_now")
@pytest.mark.usefixtures("mocked_system_query")
@pytest.mark.usefixtures("mocked_registry_fetch")
@pytest.mark.usefixtures("mocked_fetch_definition_mapping")
def test_list_packages_recorded(
    mocked_history_start_recording, mocked_history_get,
    mocked_filesystem_export, options, recorded
):
    """Record history when displaying list of available packages."""
    mocked_history_get.return_value = "__HISTORY__"
    runner = CliRunner()
    result = runner.invoke(wiz.command_line.main, options + ["list", "package"])
    print(options + ["list", "package"])
    print(result.output)
    assert result.exit_code == 0
    assert not result.exception

    if recorded:
        mocked_history_start_recording.assert_called_once_with(
            command=" ".join(["wiz"] + options + ["list", "package"])
        )
        mocked_history_get.assert_called_once_with(serialized=True)
        mocked_filesystem_export.assert_called_once_with(
            tempfile.gettempdir() + "/wiz-NOW.dump", "__HISTORY__",
            compressed=True
        )

    else:
        mocked_history_start_recording.assert_not_called()
        mocked_history_get.assert_not_called()
        mocked_filesystem_export.assert_not_called()


def test_list_packages_empty(
    mocked_system_query, mocked_registry_fetch, mocked_definition_discover
):
    """Display list of packages when no packages are available."""
    mocked_system_query.return_value = "__SYSTEM__"
    mocked_registry_fetch.return_value = []

    runner = CliRunner()
    result = runner.invoke(wiz.command_line.main, ["list", "package"])
    assert result.exit_code == 0
    assert not result.exception
    assert result.output == (
        "\n"
        "Registries\n"
        "----------\n"
        "No registries to display.\n"
        "\n"
        "\n"
        "Package   Version   Registry   Description\n"
        "-------   -------   --------   -----------\n"
        "No packages to display.\n"
        "\n"
    )

    mocked_definition_discover.assert_called_once_with(
        [], system_mapping="__SYSTEM__", max_depth=None
    )


def test_list_packages(
    mocked_system_query, mocked_registry_fetch, mocked_definition_discover,
    definitions
):
    """Display list of available packages."""
    mocked_system_query.return_value = "__SYSTEM__"
    mocked_registry_fetch.return_value = ["/registry1", "/registry2"]
    mocked_definition_discover.return_value = definitions

    runner = CliRunner()
    result = runner.invoke(
        wiz.command_line.main, ["list", "package", "--no-arch"]
    )
    assert result.exit_code == 0
    assert not result.exception
    assert result.output == (
        "\n"
        "Registries    \n"
        "--------------\n"
        "[0] /registry1\n"
        "[1] /registry2\n"
        "\n"
        "\n"
        "Package          Version   System   Registry   Description       \n"
        "--------------   -------   ------   --------   ------------------\n"
        "bar [Variant1]   0.1.0     noarch   1          This is Bar 0.1.0.\n"
        "bar [Variant2]   0.1.0     noarch   1          This is Bar 0.1.0.\n"
        "bim [Variant1]   0.1.1     noarch   1          This is Bim 0.1.1.\n"
        "foo              0.2.0     linux    0          This is Foo 0.2.0.\n"
        "foo              0.2.0     mac      0          This is Foo 0.2.0.\n"
        "\n"
    )

    mocked_definition_discover.assert_called_once_with(
        ["/registry1", "/registry2"],
        system_mapping=None, max_depth=None
    )


def test_list_packages_with_versions(
    mocked_system_query, mocked_registry_fetch, mocked_definition_discover,
    definitions
):
    """Display list of available packages with versions."""
    mocked_system_query.return_value = "__SYSTEM__"
    mocked_registry_fetch.return_value = ["/registry1", "/registry2"]
    mocked_definition_discover.return_value = definitions

    runner = CliRunner()
    result = runner.invoke(
        wiz.command_line.main, ["list", "package", "--no-arch", "--all"]
    )
    assert result.exit_code == 0
    assert not result.exception
    assert result.output == (
        "\n"
        "Registries    \n"
        "--------------\n"
        "[0] /registry1\n"
        "[1] /registry2\n"
        "\n"
        "\n"
        "Package          Version   System   Registry   Description       \n"
        "--------------   -------   ------   --------   ------------------\n"
        "bar [Variant1]   0.1.0     noarch   1          This is Bar 0.1.0.\n"
        "bar [Variant2]   0.1.0     noarch   1          This is Bar 0.1.0.\n"
        "bim [Variant1]   0.1.1     noarch   1          This is Bim 0.1.1.\n"
        "bim              0.1.0     noarch   1          This is Bim 0.1.0.\n"
        "foo              0.2.0     linux    0          This is Foo 0.2.0.\n"
        "foo              0.2.0     mac      0          This is Foo 0.2.0.\n"
        "foo              0.1.0     linux    0          This is Foo 0.1.0.\n"
        "\n"
    )

    mocked_definition_discover.assert_called_once_with(
        ["/registry1", "/registry2"],
        system_mapping=None, max_depth=None
    )


@pytest.mark.parametrize("options", [
    ["--", "--incorrect"],
    ["--incorrect"],
], ids=[
    "extra-arguments",
    "unknown-arguments",
])
@pytest.mark.usefixtures("mocked_system_query")
@pytest.mark.usefixtures("mocked_registry_fetch")
@pytest.mark.usefixtures("mocked_fetch_definition_mapping")
def test_list_packages_error(options):
    """Fail to list available packages."""
    runner = CliRunner()
    result = runner.invoke(wiz.command_line.main, ["list", "package"] + options)
    assert result.exit_code == 2
    assert result.exception


@pytest.mark.parametrize("options, recorded", [
    ([], False),
    (["--record", tempfile.gettempdir()], True)
], ids=[
    "normal",
    "recorded",
])
@pytest.mark.usefixtures("mock_datetime_now")
@pytest.mark.usefixtures("mocked_system_query")
@pytest.mark.usefixtures("mocked_registry_fetch")
@pytest.mark.usefixtures("mocked_fetch_definition_mapping")
def test_list_commands_recorded(
    mocked_history_start_recording, mocked_history_get,
    mocked_filesystem_export, options, recorded
):
    """Record history when displaying list of available commands."""
    mocked_history_get.return_value = "__HISTORY__"

    runner = CliRunner()
    result = runner.invoke(wiz.command_line.main, options + ["list", "command"])
    print(result.output)
    assert result.exit_code == 0
    assert not result.exception

    if recorded:
        mocked_history_start_recording.assert_called_once_with(
            command=" ".join(["wiz"] + options + ["list", "command"])
        )
        mocked_history_get.assert_called_once_with(serialized=True)
        mocked_filesystem_export.assert_called_once_with(
            tempfile.gettempdir() + "/wiz-NOW.dump", "__HISTORY__",
            compressed=True
        )

    else:
        mocked_history_start_recording.assert_not_called()
        mocked_history_get.assert_not_called()
        mocked_filesystem_export.assert_not_called()


def test_list_commands_empty(
    mocked_system_query, mocked_registry_fetch, mocked_definition_discover
):
    """Display list of commands when no commands are available."""
    mocked_system_query.return_value = "__SYSTEM__"
    mocked_registry_fetch.return_value = []

    runner = CliRunner()
    result = runner.invoke(wiz.command_line.main, ["list", "command"])
    assert result.exit_code == 0
    assert not result.exception
    assert result.output == (
        "\n"
        "Registries\n"
        "----------\n"
        "No registries to display.\n"
        "\n"
        "\n"
        "Command   Version   Registry   Description\n"
        "-------   -------   --------   -----------\n"
        "No commands to display.\n"
        "\n"
    )

    mocked_definition_discover.assert_called_once_with(
        [], system_mapping="__SYSTEM__", max_depth=None
    )


@pytest.mark.usefixtures("mocked_system_query")
def test_list_commands(
    mocked_system_query, mocked_registry_fetch, mocked_definition_discover,
    definitions
):
    """Display list of available commands."""
    mocked_system_query.return_value = "__SYSTEM__"
    mocked_registry_fetch.return_value = ["/registry1", "/registry2"]
    mocked_definition_discover.return_value = definitions

    runner = CliRunner()
    result = runner.invoke(
        wiz.command_line.main, ["list", "command", "--no-arch"]
    )
    assert result.exit_code == 0
    assert not result.exception
    assert result.output == (
        "\n"
        "Registries    \n"
        "--------------\n"
        "[0] /registry1\n"
        "[1] /registry2\n"
        "\n"
        "\n"
        "Command             Version   System   Registry   Description       \n"
        "-----------------   -------   ------   --------   ------------------\n"
        "bimExe [Variant1]   0.1.1     noarch   1          This is Bim 0.1.1.\n"
        "fooExe              0.2.0     linux    0          This is Foo 0.2.0.\n"
        "fooExe              0.2.0     mac      0          This is Foo 0.2.0.\n"
        "\n"
    )

    mocked_definition_discover.assert_called_once_with(
        ["/registry1", "/registry2"],
        system_mapping=None, max_depth=None
    )


def test_list_commands_with_versions(
    mocked_system_query, mocked_registry_fetch, mocked_definition_discover,
    definitions
):
    """Display list of available commands with versions."""
    mocked_system_query.return_value = "__SYSTEM__"
    mocked_registry_fetch.return_value = ["/registry1", "/registry2"]
    mocked_definition_discover.return_value = definitions

    runner = CliRunner()
    result = runner.invoke(
        wiz.command_line.main, ["list", "command", "--all", "--no-arch"]
    )
    assert result.exit_code == 0
    assert not result.exception
    assert result.output == (
        "\n"
        "Registries    \n"
        "--------------\n"
        "[0] /registry1\n"
        "[1] /registry2\n"
        "\n"
        "\n"
        "Command             Version   System   Registry   Description       \n"
        "-----------------   -------   ------   --------   ------------------\n"
        "bimExe [Variant1]   0.1.1     noarch   1          This is Bim 0.1.1.\n"
        "bimExe              0.1.0     noarch   1          This is Bim 0.1.0.\n"
        "fooExe              0.2.0     linux    0          This is Foo 0.2.0.\n"
        "fooExe              0.2.0     mac      0          This is Foo 0.2.0.\n"
        "fooExe              0.1.0     linux    0          This is Foo 0.1.0.\n"
        "\n"
    )

    mocked_definition_discover.assert_called_once_with(
        ["/registry1", "/registry2"],
        system_mapping=None, max_depth=None
    )


@pytest.mark.parametrize("options", [
    ["--", "--incorrect"],
    ["--incorrect"],
], ids=[
    "extra-arguments",
    "unknown-arguments",
])
@pytest.mark.usefixtures("mocked_system_query")
@pytest.mark.usefixtures("mocked_registry_fetch")
@pytest.mark.usefixtures("mocked_fetch_definition_mapping")
def test_list_commands_error(options):
    """Fail to list available commands."""
    runner = CliRunner()
    result = runner.invoke(wiz.command_line.main, ["list", "command"] + options)
    assert result.exit_code == 2
    assert result.exception


@pytest.mark.parametrize("options, recorded", [
    ([], False),
    (["--record", tempfile.gettempdir()], True)
], ids=[
    "normal",
    "recorded",
])
@pytest.mark.usefixtures("mock_datetime_now")
@pytest.mark.usefixtures("mocked_system_query")
@pytest.mark.usefixtures("mocked_registry_fetch")
@pytest.mark.usefixtures("mocked_fetch_definition_mapping")
def test_search_recorded(
    mocked_history_start_recording, mocked_history_get,
    mocked_filesystem_export, options, recorded
):
    """Record history when searching packages and commands."""
    mocked_history_get.return_value = "__HISTORY__"

    runner = CliRunner()
    result = runner.invoke(wiz.command_line.main, options + ["search", "foo"])
    assert result.exit_code == 0
    assert not result.exception

    if recorded:
        mocked_history_start_recording.assert_called_once_with(
            command=" ".join(["wiz"] + options + ["search", "foo"])
        )
        mocked_history_get.assert_called_once_with(serialized=True)
        mocked_filesystem_export.assert_called_once_with(
            tempfile.gettempdir() + "/wiz-NOW.dump", "__HISTORY__",
            compressed=True
        )

    else:
        mocked_history_start_recording.assert_not_called()
        mocked_history_get.assert_not_called()
        mocked_filesystem_export.assert_not_called()


def test_search_empty(
    mocked_system_query, mocked_registry_fetch, mocked_definition_discover,
    logger
):
    """Display empty list of searched commands and packages."""
    mocked_system_query.return_value = "__SYSTEM__"
    mocked_registry_fetch.return_value = []

    runner = CliRunner()
    result = runner.invoke(wiz.command_line.main, ["search", "foo"])
    assert result.exit_code == 0
    assert not result.exception
    assert result.output == (
        "\n"
        "Registries\n"
        "----------\n"
        "No registries to display.\n"
        "\n"
    )

    logger.warning.assert_called_once_with("No results found.\n")

    mocked_definition_discover.assert_called_once_with(
        [], system_mapping="__SYSTEM__", max_depth=None
    )


@pytest.mark.parametrize("options", [
    [], ["--type", "all"]
], ids=[
    "default",
    "with-option",
])
def test_search(
    mocked_system_query, mocked_registry_fetch, mocked_definition_discover,
    logger, definitions, options
):
    """Display searched available commands and packages."""
    mocked_system_query.return_value = "__SYSTEM__"
    mocked_registry_fetch.return_value = ["/registry1", "/registry2"]
    mocked_definition_discover.return_value = definitions

    runner = CliRunner()
    result = runner.invoke(
        wiz.command_line.main, ["search", "bim"] + options
    )
    assert not result.exception
    assert result.exit_code == 0
    assert result.output == (
        "\n"
        "Registries    \n"
        "--------------\n"
        "[0] /registry1\n"
        "[1] /registry2\n"
        "\n"
        "\n"
        "Command             Version   Registry   Description       \n"
        "-----------------   -------   --------   ------------------\n"
        "bimExe [Variant1]   0.1.1     1          This is Bim 0.1.1.\n"
        "\n"
        "\n"
        "Package          Version   Registry   Description       \n"
        "--------------   -------   --------   ------------------\n"
        "bim [Variant1]   0.1.1     1          This is Bim 0.1.1.\n"
        "\n"
    )

    logger.warning.assert_not_called()

    mocked_definition_discover.assert_called_once_with(
        ["/registry1", "/registry2"],
        system_mapping="__SYSTEM__",
        max_depth=None
    )


@pytest.mark.parametrize("options", [
    [], ["--type", "all"]
], ids=[
    "default",
    "with-option",
])
def test_search_filtered_command(
    mocked_system_query, mocked_registry_fetch, mocked_definition_discover,
    logger, definitions, options
):
    """Display searched available commands and packages."""
    mocked_system_query.return_value = "__SYSTEM__"
    mocked_registry_fetch.return_value = ["/registry1", "/registry2"]
    mocked_definition_discover.return_value = definitions

    runner = CliRunner()
    result = runner.invoke(
        wiz.command_line.main, ["search", "Exe", "--no-arch"] + options
    )
    assert result.exit_code == 0
    assert not result.exception
    assert result.output == (
        "\n"
        "Registries    \n"
        "--------------\n"
        "[0] /registry1\n"
        "[1] /registry2\n"
        "\n"
        "\n"
        "Command             Version   System   Registry   Description       \n"
        "-----------------   -------   ------   --------   ------------------\n"
        "bimExe [Variant1]   0.1.1     noarch   1          This is Bim 0.1.1.\n"
        "fooExe              0.2.0     linux    0          This is Foo 0.2.0.\n"
        "fooExe              0.2.0     mac      0          This is Foo 0.2.0.\n"
        "\n"
        "\n"
        "Package          Version   System   Registry   Description       \n"
        "--------------   -------   ------   --------   ------------------\n"
        "bim [Variant1]   0.1.1     noarch   1          This is Bim 0.1.1.\n"
        "foo              0.2.0     linux    0          This is Foo 0.2.0.\n"
        "foo              0.2.0     mac      0          This is Foo 0.2.0.\n"
        "\n"
    )

    logger.warning.assert_not_called()

    mocked_definition_discover.assert_called_once_with(
        ["/registry1", "/registry2"],
        system_mapping=None,
        max_depth=None
    )


def test_search_with_versions(
    mocked_system_query, mocked_registry_fetch, mocked_definition_discover,
    logger, definitions
):
    """Display searched commands and packages with all versions."""
    mocked_system_query.return_value = "__SYSTEM__"
    mocked_registry_fetch.return_value = ["/registry1", "/registry2"]
    mocked_definition_discover.return_value = definitions

    runner = CliRunner()
    result = runner.invoke(
        wiz.command_line.main, ["search", "bim", "--all", "--no-arch"]
    )
    assert result.exit_code == 0
    assert not result.exception
    assert result.output == (
        "\n"
        "Registries    \n"
        "--------------\n"
        "[0] /registry1\n"
        "[1] /registry2\n"
        "\n"
        "\n"
        "Command             Version   System   Registry   Description       \n"
        "-----------------   -------   ------   --------   ------------------\n"
        "bimExe [Variant1]   0.1.1     noarch   1          This is Bim 0.1.1.\n"
        "bimExe              0.1.0     noarch   1          This is Bim 0.1.0.\n"
        "\n"
        "\n"
        "Package          Version   System   Registry   Description       \n"
        "--------------   -------   ------   --------   ------------------\n"
        "bim [Variant1]   0.1.1     noarch   1          This is Bim 0.1.1.\n"
        "bim              0.1.0     noarch   1          This is Bim 0.1.0.\n"
        "\n"
    )

    logger.warning.assert_not_called()

    mocked_definition_discover.assert_called_once_with(
        ["/registry1", "/registry2"],
        system_mapping=None,
        max_depth=None
    )


def test_search_packages(
    mocked_system_query, mocked_registry_fetch, mocked_definition_discover,
    definitions
):
    """Display searched list of available packages."""
    mocked_system_query.return_value = "__SYSTEM__"
    mocked_registry_fetch.return_value = ["/registry1", "/registry2"]
    mocked_definition_discover.return_value = definitions

    runner = CliRunner()
    result = runner.invoke(
        wiz.command_line.main, ["search", "bim", "-t", "package"]
    )
    assert result.exit_code == 0
    assert not result.exception
    assert result.output == (
        "\n"
        "Registries    \n"
        "--------------\n"
        "[0] /registry1\n"
        "[1] /registry2\n"
        "\n"
        "\n"
        "Package          Version   Registry   Description       \n"
        "--------------   -------   --------   ------------------\n"
        "bim [Variant1]   0.1.1     1          This is Bim 0.1.1.\n"
        "\n"
    )

    mocked_definition_discover.assert_called_once_with(
        ["/registry1", "/registry2"],
        system_mapping="__SYSTEM__",
        max_depth=None
    )


def test_search_packages_with_versions(
    mocked_system_query, mocked_registry_fetch, mocked_definition_discover,
    definitions
):
    """Display searched available packages with all versions."""
    mocked_system_query.return_value = "__SYSTEM__"
    mocked_registry_fetch.return_value = ["/registry1", "/registry2"]
    mocked_definition_discover.return_value = definitions

    runner = CliRunner()
    result = runner.invoke(
        wiz.command_line.main,
        ["search", "bim", "-t", "package", "--all"]
    )
    assert result.exit_code == 0
    assert not result.exception
    assert result.output == (
        "\n"
        "Registries    \n"
        "--------------\n"
        "[0] /registry1\n"
        "[1] /registry2\n"
        "\n"
        "\n"
        "Package          Version   Registry   Description       \n"
        "--------------   -------   --------   ------------------\n"
        "bim [Variant1]   0.1.1     1          This is Bim 0.1.1.\n"
        "bim              0.1.0     1          This is Bim 0.1.0.\n"
        "\n"
    )

    mocked_definition_discover.assert_called_once_with(
        ["/registry1", "/registry2"],
        system_mapping="__SYSTEM__",
        max_depth=None
    )


def test_search_commands(
    mocked_system_query, mocked_registry_fetch, mocked_definition_discover,
    definitions
):
    """Display searched list of available commands."""
    mocked_system_query.return_value = "__SYSTEM__"
    mocked_registry_fetch.return_value = ["/registry1", "/registry2"]
    mocked_definition_discover.return_value = definitions

    runner = CliRunner()
    result = runner.invoke(
        wiz.command_line.main, ["search", "bim", "-t", "command"]
    )
    assert result.exit_code == 0
    assert not result.exception
    assert result.output == (
        "\n"
        "Registries    \n"
        "--------------\n"
        "[0] /registry1\n"
        "[1] /registry2\n"
        "\n"
        "\n"
        "Command             Version   Registry   Description       \n"
        "-----------------   -------   --------   ------------------\n"
        "bimExe [Variant1]   0.1.1     1          This is Bim 0.1.1.\n"
        "\n"
    )

    mocked_definition_discover.assert_called_once_with(
        ["/registry1", "/registry2"],
        system_mapping="__SYSTEM__",
        max_depth=None
    )


def test_search_commands_with_versions(
    mocked_system_query, mocked_registry_fetch, mocked_definition_discover,
    definitions
):
    """Display searched available commands with all versions."""
    mocked_system_query.return_value = "__SYSTEM__"
    mocked_registry_fetch.return_value = ["/registry1", "/registry2"]
    mocked_definition_discover.return_value = definitions

    runner = CliRunner()
    result = runner.invoke(
        wiz.command_line.main,
        ["search", "bim", "-t", "command", "--all"]
    )
    assert result.exit_code == 0
    assert not result.exception
    assert result.output == (
        "\n"
        "Registries    \n"
        "--------------\n"
        "[0] /registry1\n"
        "[1] /registry2\n"
        "\n"
        "\n"
        "Command             Version   Registry   Description       \n"
        "-----------------   -------   --------   ------------------\n"
        "bimExe [Variant1]   0.1.1     1          This is Bim 0.1.1.\n"
        "bimExe              0.1.0     1          This is Bim 0.1.0.\n"
        "\n"
    )

    mocked_definition_discover.assert_called_once_with(
        ["/registry1", "/registry2"],
        system_mapping="__SYSTEM__",
        max_depth=None
    )


@pytest.mark.parametrize("options", [
    ["--", "--incorrect"],
    ["--incorrect"],
], ids=[
    "extra-arguments",
    "unknown-arguments",
])
@pytest.mark.usefixtures("mocked_system_query")
@pytest.mark.usefixtures("mocked_registry_fetch")
@pytest.mark.usefixtures("mocked_fetch_definition_mapping")
def test_search_error(options):
    """Fail to search available packages and commands."""
    runner = CliRunner()
    result = runner.invoke(wiz.command_line.main, ["search", "foo"] + options)
    assert result.exit_code == 2
    assert result.exception


@pytest.mark.parametrize("options, recorded", [
    ([], False),
    (["--record", tempfile.gettempdir()], True)
], ids=[
    "normal",
    "recorded",
])
@pytest.mark.usefixtures("mock_datetime_now")
@pytest.mark.usefixtures("mocked_system_query")
@pytest.mark.usefixtures("mocked_registry_fetch")
@pytest.mark.usefixtures("mocked_fetch_definition_mapping")
def test_view_recorded(
    mocked_history_start_recording, mocked_history_get,
    mocked_filesystem_export, options, recorded
):
    """Record history when viewing a definition."""
    mocked_history_get.return_value = "__HISTORY__"

    runner = CliRunner()
    result = runner.invoke(wiz.command_line.main, options + ["view", "foo"])
    assert result.exit_code == 0
    assert not result.exception

    if recorded:
        mocked_history_start_recording.assert_called_once_with(
            command=" ".join(["wiz"] + options + ["view", "foo"])
        )
        mocked_history_get.assert_called_once_with(serialized=True)
        mocked_filesystem_export.assert_called_once_with(
            tempfile.gettempdir() + "/wiz-NOW.dump", "__HISTORY__",
            compressed=True
        )

    else:
        mocked_history_start_recording.assert_not_called()
        mocked_history_get.assert_not_called()
        mocked_filesystem_export.assert_not_called()


def test_view_not_found(
    mocked_system_query, mocked_registry_fetch, mocked_fetch_definition_mapping,
    logger
):
    """Fail to view definition when request cannot be found."""
    mocked_system_query.return_value = "__SYSTEM__"
    mocked_registry_fetch.return_value = ["/registry1", "/registry2"]
    mocked_fetch_definition_mapping.return_value = {
        "command": {},
        "package": {}
    }

    runner = CliRunner()
    result = runner.invoke(wiz.command_line.main, ["view", "foo"])
    assert result.exit_code == 0
    assert not result.exception
    assert result.output == ""

    logger.warning.assert_called_once_with("No definition found.\n")

    mocked_fetch_definition_mapping.assert_called_once_with(
        ["/registry1", "/registry2"],
        system_mapping="__SYSTEM__", max_depth=None
    )


def test_view_definition(
    mocked_system_query, mocked_registry_fetch, mocked_fetch_definition_mapping,
    definition_mapping, logger
):
    """Display definition from identifier."""
    mocked_system_query.return_value = "__SYSTEM__"
    mocked_registry_fetch.return_value = ["/registry1", "/registry2"]
    mocked_fetch_definition_mapping.return_value = definition_mapping

    runner = CliRunner()
    result = runner.invoke(wiz.command_line.main, ["view", "bar"])
    assert not result.exception
    assert result.exit_code == 0
    assert result.output == (
        "path: /registry1/bar-0.1.0.json\n"
        "registry: /registry2\n"
        "identifier: bar\n"
        "version: 0.1.0\n"
        "description: This is Bar 0.1.0.\n"
        "environ:\n"
        "    PATH: /path/to/bin:${PATH}\n"
        "requirements:\n"
        "    bim >= 0.1.0, < 1\n"
        "variants:\n"
        "    identifier: Variant1\n"
        "        environ:\n"
        "            PYTHONPATH: /path/to/lib/1:${PYTHONPATH}\n"
        "    identifier: Variant2\n"
        "        environ:\n"
        "            PYTHONPATH: /path/to/lib/2:${PYTHONPATH}\n"
    )

    logger.info.assert_called_once_with("View definition: bar==0.1.0")
    logger.warning.assert_not_called()

    mocked_fetch_definition_mapping.assert_called_once_with(
        ["/registry1", "/registry2"],
        system_mapping="__SYSTEM__", max_depth=None
    )


def test_view_definition_json(
    mocked_system_query, mocked_registry_fetch, mocked_fetch_definition_mapping,
    definition_mapping, logger
):
    """Display definition from identifier in JSON format."""
    mocked_system_query.return_value = "__SYSTEM__"
    mocked_registry_fetch.return_value = ["/registry1", "/registry2"]
    mocked_fetch_definition_mapping.return_value = definition_mapping

    runner = CliRunner()
    result = runner.invoke(wiz.command_line.main, ["view", "bar", "--json"])
    assert not result.exception
    assert result.exit_code == 0
    assert result.output == (
        "{\n"
        "    \"identifier\": \"bar\",\n"
        "    \"version\": \"0.1.0\",\n"
        "    \"description\": \"This is Bar 0.1.0.\",\n"
        "    \"environ\": {\n"
        "        \"PATH\": \"/path/to/bin:${PATH}\"\n"
        "    },\n"
        "    \"requirements\": [\n"
        "        \"bim >= 0.1.0, < 1\"\n"
        "    ],\n"
        "    \"variants\": [\n"
        "        {\n"
        "            \"identifier\": \"Variant1\",\n"
        "            \"environ\": {\n"
        "                \"PYTHONPATH\": \"/path/to/lib/1:${PYTHONPATH}\"\n"
        "            }\n"
        "        },\n"
        "        {\n"
        "            \"identifier\": \"Variant2\",\n"
        "            \"environ\": {\n"
        "                \"PYTHONPATH\": \"/path/to/lib/2:${PYTHONPATH}\"\n"
        "            }\n"
        "        }\n"
        "    ]\n"
        "}\n"
    )

    logger.info.assert_called_once_with("View definition: bar==0.1.0")
    logger.warning.assert_not_called()

    mocked_fetch_definition_mapping.assert_called_once_with(
        ["/registry1", "/registry2"],
        system_mapping="__SYSTEM__", max_depth=None
    )


def test_view_definition_from_command(
    mocked_system_query, mocked_registry_fetch, mocked_fetch_definition_mapping,
    definition_mapping, logger
):
    """Indicate that identifier is referring to a definition."""
    mocked_system_query.return_value = "__SYSTEM__"
    mocked_registry_fetch.return_value = ["/registry1", "/registry2"]
    mocked_fetch_definition_mapping.return_value = definition_mapping

    runner = CliRunner()
    result = runner.invoke(wiz.command_line.main, ["view", "fooExe"])
    assert result.exit_code == 0
    assert not result.exception
    assert result.output == ""

    logger.info.assert_called_once_with(
        "Command found in definition: foo==0.2.0"
    )
    logger.warning.assert_not_called()

    mocked_fetch_definition_mapping.assert_called_once_with(
        ["/registry1", "/registry2"],
        system_mapping="__SYSTEM__", max_depth=None
    )


@pytest.mark.parametrize("options", [
    ["--", "--incorrect"],
    ["--incorrect"],
], ids=[
    "extra-arguments",
    "unknown-arguments",
])
@pytest.mark.usefixtures("mocked_system_query")
@pytest.mark.usefixtures("mocked_registry_fetch")
@pytest.mark.usefixtures("mocked_fetch_definition_mapping")
def test_view_error(options):
    """Fail to view definitions."""
    runner = CliRunner()
    result = runner.invoke(wiz.command_line.main, ["view", "foo"] + options)
    assert result.exit_code == 2
    assert result.exception


@pytest.mark.parametrize("options, recorded", [
    ([], False),
    (["--record", tempfile.gettempdir()], True)
], ids=[
    "normal",
    "recorded",
])
@pytest.mark.usefixtures("mock_datetime_now")
@pytest.mark.usefixtures("mocked_system_query")
@pytest.mark.usefixtures("mocked_registry_fetch")
@pytest.mark.usefixtures("mocked_fetch_definition_mapping")
@pytest.mark.usefixtures("mocked_resolve_context")
@pytest.mark.usefixtures("mocked_resolve_command")
@pytest.mark.usefixtures("mocked_spawn_execute")
@pytest.mark.usefixtures("mocked_spawn_shell")
@pytest.mark.usefixtures("mocked_history_record_action")
def test_use_recorded(
    mocked_history_start_recording, mocked_history_get,
    mocked_filesystem_export, options, recorded
):
    """Record history when using a resolved context."""
    mocked_history_get.return_value = "__HISTORY__"

    runner = CliRunner()
    result = runner.invoke(wiz.command_line.main, options + ["use", "foo"])
    assert result.exit_code == 0
    assert not result.exception

    if recorded:
        mocked_history_start_recording.assert_called_once_with(
            command=" ".join(["wiz"] + options + ["use", "foo"])
        )
        mocked_history_get.assert_called_once_with(serialized=True)
        mocked_filesystem_export.assert_called_once_with(
            tempfile.gettempdir() + "/wiz-NOW.dump", "__HISTORY__",
            compressed=True
        )

    else:
        mocked_history_start_recording.assert_not_called()
        mocked_history_get.assert_not_called()
        mocked_filesystem_export.assert_not_called()


@pytest.mark.parametrize("options, max_combinations, max_attempts", [
    ([], 10, 15),
    (["-mc", "1"], 1, 15),
    (["-ma", "1"], 10, 1),
], ids=[
    "simple",
    "with-maximum-combinations",
    "with-maximum-attempts",
])
def test_use_spawn_shell(
    mocked_system_query, mocked_registry_fetch, mocked_fetch_definition_mapping,
    mocked_resolve_context, mocked_resolve_command, mocked_spawn_execute,
    mocked_spawn_shell, mocked_history_record_action, wiz_context, logger,
    options, max_combinations, max_attempts
):
    """Use a resolved context."""
    mocked_system_query.return_value = "__SYSTEM__"
    mocked_registry_fetch.return_value = ["/registry1", "/registry2"]
    mocked_fetch_definition_mapping.return_value = "__MAPPING__"
    mocked_resolve_context.return_value = wiz_context

    runner = CliRunner()
    result = runner.invoke(wiz.command_line.main, ["use", "foo"] + options)
    assert result.output == ""
    assert result.exit_code == 0
    assert not result.exception

    mocked_fetch_definition_mapping.assert_called_once_with(
        ["/registry1", "/registry2"],
        system_mapping="__SYSTEM__", max_depth=None
    )

    mocked_resolve_context.assert_called_once_with(
        ["foo"], "__MAPPING__",
        ignore_implicit=False,
        environ_mapping={},
        maximum_combinations=max_combinations,
        maximum_attempts=max_attempts,
    )

    mocked_spawn_shell.assert_called_once_with({
        "KEY1": "value1",
        "KEY2": "value2"
    }, {
        "fooExe": "foo",
        "fooExeDebug": "foo --debug",
    })

    mocked_resolve_command.assert_not_called()
    mocked_spawn_execute.assert_not_called()
    mocked_history_record_action.assert_not_called()
    logger.error.assert_not_called()


@pytest.mark.parametrize("options, max_combinations, max_attempts", [
    ([], 10, 15),
    (["-mc", "1"], 1, 15),
    (["-ma", "1"], 10, 1),
], ids=[
    "simple",
    "with-maximum-combinations",
    "with-maximum-attempts",
])
def test_use_spawn_shell_view(
    mocked_system_query, mocked_registry_fetch, mocked_fetch_definition_mapping,
    mocked_resolve_context, mocked_resolve_command, mocked_spawn_execute,
    mocked_spawn_shell, mocked_history_record_action, wiz_context, logger,
    options, max_combinations, max_attempts
):
    """View a resolved context."""
    mocked_system_query.return_value = "__SYSTEM__"
    mocked_registry_fetch.return_value = ["/registry1", "/registry2"]
    mocked_fetch_definition_mapping.return_value = "__MAPPING__"
    mocked_resolve_context.return_value = wiz_context

    runner = CliRunner()
    result = runner.invoke(
        wiz.command_line.main, ["use", "foo", "--view"] + options
    )
    assert not result.exception
    assert result.exit_code == 0
    assert result.output == (
        "\n"
        "Registries    \n"
        "--------------\n"
        "[0] /registry1\n"
        "[1] /registry2\n"
        "\n"
        "\n"
        "Package          Version   Registry   Description       \n"
        "--------------   -------   --------   ------------------\n"
        "foo              0.1.0     0          This is Foo 0.1.0.\n"
        "bim [Variant1]   0.1.1     1          This is Bim 0.1.1.\n"
        "\n"
        "\n"
        "Command       Value      \n"
        "-----------   -----------\n"
        "fooExe        foo        \n"
        "fooExeDebug   foo --debug\n"
        "\n"
        "\n"
        "Environment Variable   Environment Value\n"
        "--------------------   -----------------\n"
        "KEY1                   value1           \n"
        "KEY2                   value2           \n"
        "\n"
    )

    mocked_fetch_definition_mapping.assert_called_once_with(
        ["/registry1", "/registry2"],
        system_mapping="__SYSTEM__", max_depth=None
    )

    mocked_resolve_context.assert_called_once_with(
        ["foo"], "__MAPPING__",
        ignore_implicit=False,
        environ_mapping={},
        maximum_combinations=max_combinations,
        maximum_attempts=max_attempts,
    )

    mocked_spawn_shell.assert_not_called()
    mocked_resolve_command.assert_not_called()
    mocked_spawn_execute.assert_not_called()
    mocked_history_record_action.assert_not_called()
    logger.error.assert_not_called()


@pytest.mark.parametrize("options, max_combinations, max_attempts", [
    ([], 10, 15),
    (["-mc", "1"], 1, 15),
    (["-ma", "1"], 10, 1),
], ids=[
    "simple",
    "with-maximum-combinations",
    "with-maximum-attempts",
])
def test_use_spawn_shell_view_empty(
    mocked_system_query, mocked_registry_fetch, mocked_fetch_definition_mapping,
    mocked_resolve_context, mocked_resolve_command, mocked_spawn_execute,
    mocked_spawn_shell, mocked_history_record_action, logger,
    options, max_combinations, max_attempts
):
    """View an empty resolved context."""
    mocked_system_query.return_value = "__SYSTEM__"
    mocked_registry_fetch.return_value = ["/registry1", "/registry2"]
    mocked_fetch_definition_mapping.return_value = "__MAPPING__"
    mocked_resolve_context.return_value = {
        "command": {},
        "environ": {},
        "packages": [],
        "registries": ["/registry1", "/registry2"]
    }

    runner = CliRunner()
    result = runner.invoke(
        wiz.command_line.main, ["use", "foo", "--view"] + options
    )
    assert result.exit_code == 0
    assert not result.exception
    assert result.output == (
        "\n"
        "Registries    \n"
        "--------------\n"
        "[0] /registry1\n"
        "[1] /registry2\n"
        "\n"
        "\n"
        "Package   Version   Registry   Description\n"
        "-------   -------   --------   -----------\n"
        "No packages to display.\n"
        "\n"
        "\n"
        "Command   Value\n"
        "-------   -----\n"
        "No commands to display.\n"
        "\n"
        "\n"
        "Environment Variable   Environment Value\n"
        "--------------------   -----------------\n"
        "No environment variables to display.\n"
        "\n"
    )

    mocked_fetch_definition_mapping.assert_called_once_with(
        ["/registry1", "/registry2"],
        system_mapping="__SYSTEM__", max_depth=None
    )

    mocked_resolve_context.assert_called_once_with(
        ["foo"], "__MAPPING__",
        ignore_implicit=False,
        environ_mapping={},
        maximum_combinations=max_combinations,
        maximum_attempts=max_attempts,
    )

    mocked_spawn_shell.assert_not_called()
    mocked_resolve_command.assert_not_called()
    mocked_spawn_execute.assert_not_called()
    mocked_history_record_action.assert_not_called()
    logger.error.assert_not_called()


@pytest.mark.parametrize("options, max_combinations, max_attempts", [
    ([], 10, 15),
    (["-mc", "1"], 1, 15),
    (["-ma", "1"], 10, 1),
], ids=[
    "simple",
    "with-maximum-combinations",
    "with-maximum-attempts",
])
def test_use_execute_command(
    mocked_system_query, mocked_registry_fetch, mocked_fetch_definition_mapping,
    mocked_resolve_context, mocked_resolve_command, mocked_spawn_execute,
    mocked_spawn_shell, mocked_history_record_action, wiz_context, logger,
    options, max_combinations, max_attempts
):
    """Execute a command within a resolved context."""
    mocked_system_query.return_value = "__SYSTEM__"
    mocked_registry_fetch.return_value = ["/registry1", "/registry2"]
    mocked_fetch_definition_mapping.return_value = "__MAPPING__"
    mocked_resolve_context.return_value = wiz_context
    mocked_resolve_command.return_value = "__RESOLVED_COMMAND__"

    runner = CliRunner()
    result = runner.invoke(
        wiz.command_line.main,
        ["use", "foo"] + options + [
            "--", "fooExeDebug", "-t", "/path/to/script.foo"
        ]
    )
    assert result.exit_code == 0
    assert not result.exception
    assert result.output == ""

    mocked_fetch_definition_mapping.assert_called_once_with(
        ["/registry1", "/registry2"],
        system_mapping="__SYSTEM__", max_depth=None
    )

    mocked_resolve_context.assert_called_once_with(
        ["foo"], "__MAPPING__",
        ignore_implicit=False,
        environ_mapping={},
        maximum_combinations=max_combinations,
        maximum_attempts=max_attempts,
    )

    mocked_resolve_command.assert_called_once_with(
        ["fooExeDebug", "-t", "/path/to/script.foo"],
        {
            "fooExe": "foo",
            "fooExeDebug": "foo --debug",
        }
    )

    mocked_spawn_execute.assert_called_once_with(
        "__RESOLVED_COMMAND__",
        {
            "KEY1": "value1",
            "KEY2": "value2"
        }
    )

    mocked_spawn_shell.assert_not_called()
    mocked_history_record_action.assert_not_called()
    logger.error.assert_not_called()


@pytest.mark.parametrize("options, max_combinations, max_attempts", [
    ([], 10, 15),
    (["-mc", "1"], 1, 15),
    (["-ma", "1"], 10, 1),
], ids=[
    "simple",
    "with-maximum-combinations",
    "with-maximum-attempts",
])
def test_use_with_resolution_error(
    mocked_system_query, mocked_registry_fetch, mocked_fetch_definition_mapping,
    mocked_resolve_context, mocked_resolve_command, mocked_spawn_execute,
    mocked_spawn_shell, mocked_history_record_action, logger,
    options, max_combinations, max_attempts
):
    """Fail to resolve a context."""
    exception = wiz.exception.WizError("Oh Shit!")
    mocked_system_query.return_value = "__SYSTEM__"
    mocked_registry_fetch.return_value = ["/registry1", "/registry2"]
    mocked_fetch_definition_mapping.return_value = "__MAPPING__"
    mocked_resolve_context.side_effect = exception

    runner = CliRunner()
    result = runner.invoke(
        wiz.command_line.main, ["use", "foo", "bim==0.1.*"] + options
    )
    assert result.exit_code == 0
    assert not result.exception
    assert result.output == ""

    mocked_fetch_definition_mapping.assert_called_once_with(
        ["/registry1", "/registry2"],
        system_mapping="__SYSTEM__", max_depth=None
    )

    mocked_resolve_context.assert_called_once_with(
        ["foo", "bim==0.1.*"], "__MAPPING__",
        ignore_implicit=False,
        environ_mapping={},
        maximum_combinations=max_combinations,
        maximum_attempts=max_attempts,
    )

    mocked_spawn_shell.assert_not_called()
    mocked_resolve_command.assert_not_called()
    mocked_spawn_execute.assert_not_called()

    mocked_history_record_action.assert_called_once_with(
        "RAISE_EXCEPTION", error=exception
    )

    logger.error.assert_called_once_with("Oh Shit!")


@pytest.mark.parametrize("options", [
    ["--incorrect"],
], ids=[
    "unknown-arguments",
])
@pytest.mark.usefixtures("mocked_system_query")
@pytest.mark.usefixtures("mocked_registry_fetch")
@pytest.mark.usefixtures("mocked_fetch_definition_mapping")
@pytest.mark.usefixtures("mocked_resolve_context")
@pytest.mark.usefixtures("mocked_resolve_command")
@pytest.mark.usefixtures("mocked_spawn_execute")
@pytest.mark.usefixtures("mocked_spawn_shell")
@pytest.mark.usefixtures("mocked_history_record_action")
def test_use_error(options):
    """Fail to view definitions."""
    runner = CliRunner()
    result = runner.invoke(wiz.command_line.main, ["use", "foo"] + options)
    assert result.exit_code == 2
    assert result.exception


@pytest.mark.parametrize("options, max_combinations, max_attempts", [
    ([], 10, 15),
    (["-mc", "1"], 1, 15),
    (["-ma", "1"], 10, 1),
], ids=[
    "simple",
    "with-maximum-combinations",
    "with-maximum-attempts",
])
def test_use_initial_environment(
    mocked_system_query, mocked_registry_fetch, mocked_fetch_definition_mapping,
    mocked_resolve_context, mocked_resolve_command, mocked_spawn_execute,
    wiz_context, options, max_combinations, max_attempts
):
    """Executing a command to extend an initial environment."""
    mocked_system_query.return_value = "__SYSTEM__"
    mocked_registry_fetch.return_value = ["/registry1", "/registry2"]
    mocked_fetch_definition_mapping.return_value = "__MAPPING__"
    mocked_resolve_context.return_value = wiz_context
    mocked_resolve_command.return_value = "__RESOLVED_COMMAND__"

    runner = CliRunner()
    result = runner.invoke(
        wiz.command_line.main,
        [
            "--init", "PATH=/path", "--init", "PYTHONPATH=/other-path",
            "use", "foo"
        ] + options + ["--", "fooExeDebug", "-t", "/path/to/script.foo"]
    )
    assert result.exit_code == 0
    assert not result.exception
    assert result.output == ""

    mocked_resolve_context.assert_called_once_with(
        ["foo"], "__MAPPING__", ignore_implicit=False,
        environ_mapping={"PATH": "/path", "PYTHONPATH": "/other-path"},
        maximum_combinations=max_combinations,
        maximum_attempts=max_attempts,
    )

    mocked_spawn_execute.assert_called_once_with(
        "__RESOLVED_COMMAND__", {"KEY1": "value1", "KEY2": "value2"}
    )


@pytest.mark.parametrize("options, recorded", [
    ([], False),
    (["--record", tempfile.gettempdir()], True)
], ids=[
    "normal",
    "recorded",
])
@pytest.mark.usefixtures("mock_datetime_now")
@pytest.mark.usefixtures("mocked_system_query")
@pytest.mark.usefixtures("mocked_registry_fetch")
@pytest.mark.usefixtures("mocked_fetch_definition_mapping")
@pytest.mark.usefixtures("mocked_fetch_package_request_from_command")
@pytest.mark.usefixtures("mocked_resolve_context")
@pytest.mark.usefixtures("mocked_spawn_execute")
@pytest.mark.usefixtures("mocked_history_record_action")
def test_run_recorded(
    mocked_history_start_recording, mocked_history_get,
    mocked_filesystem_export, mocked_resolve_command, options, recorded
):
    """Record history when running command within a resolved context."""
    mocked_history_get.return_value = "__HISTORY__"
    mocked_resolve_command.return_value = "__RESOLVED_COMMAND__"

    runner = CliRunner()
    result = runner.invoke(wiz.command_line.main, options + ["run", "fooExe"])
    assert result.exit_code == 0
    assert not result.exception

    if recorded:
        mocked_history_start_recording.assert_called_once_with(
            command=" ".join(["wiz"] + options + ["run", "fooExe"])
        )
        mocked_history_get.assert_called_once_with(serialized=True)
        mocked_filesystem_export.assert_called_once_with(
            tempfile.gettempdir() + "/wiz-NOW.dump", "__HISTORY__",
            compressed=True
        )

    else:
        mocked_history_start_recording.assert_not_called()
        mocked_history_get.assert_not_called()
        mocked_filesystem_export.assert_not_called()


@pytest.mark.parametrize("options, max_combinations, max_attempts", [
    ([], 10, 15),
    (["-mc", "1"], 1, 15),
    (["-ma", "1"], 10, 1),
], ids=[
    "simple",
    "with-maximum-combinations",
    "with-maximum-attempts",
])
def test_run(
    mocked_system_query, mocked_registry_fetch, mocked_fetch_definition_mapping,
    mocked_fetch_package_request_from_command, mocked_resolve_context,
    mocked_resolve_command, mocked_spawn_execute, mocked_history_record_action,
    wiz_context, logger, options, max_combinations, max_attempts
):
    """Execute a command within a resolved context."""
    mocked_system_query.return_value = "__SYSTEM__"
    mocked_registry_fetch.return_value = ["/registry1", "/registry2"]
    mocked_fetch_definition_mapping.return_value = "__MAPPING__"
    mocked_resolve_context.return_value = wiz_context
    mocked_fetch_package_request_from_command.return_value = "__PACKAGE__"
    mocked_resolve_command.return_value = "__RESOLVED_COMMAND__"

    runner = CliRunner()
    result = runner.invoke(wiz.command_line.main, ["run", "fooExe"] + options)
    assert result.exit_code == 0
    assert not result.exception
    assert result.output == ""

    mocked_fetch_definition_mapping.assert_called_once_with(
        ["/registry1", "/registry2"],
        system_mapping="__SYSTEM__", max_depth=None
    )

    mocked_fetch_package_request_from_command.assert_called_once_with(
        "fooExe", "__MAPPING__"
    )

    mocked_resolve_context.assert_called_once_with(
        ["__PACKAGE__"], "__MAPPING__",
        ignore_implicit=False,
        environ_mapping={},
        maximum_combinations=max_combinations,
        maximum_attempts=max_attempts,
    )

    mocked_resolve_command.assert_called_once_with(
        ["fooExe"],
        {
            "fooExe": "foo",
            "fooExeDebug": "foo --debug",
        }
    )

    mocked_spawn_execute.assert_called_once_with(
        "__RESOLVED_COMMAND__",
        {
            "KEY1": "value1",
            "KEY2": "value2"
        }
    )

    mocked_history_record_action.assert_not_called()
    logger.error.assert_not_called()


@pytest.mark.parametrize("options, max_combinations, max_attempts", [
    ([], 10, 15),
    (["-mc", "1"], 1, 15),
    (["-ma", "1"], 10, 1),
], ids=[
    "simple",
    "with-maximum-combinations",
    "with-maximum-attempts",
])
def test_run_view(
    mocked_system_query, mocked_registry_fetch, mocked_fetch_definition_mapping,
    mocked_fetch_package_request_from_command, mocked_resolve_context,
    mocked_resolve_command, mocked_spawn_execute, mocked_history_record_action,
    wiz_context, logger, options, max_combinations, max_attempts
):
    """View a resolved context from a command execution."""
    mocked_system_query.return_value = "__SYSTEM__"
    mocked_registry_fetch.return_value = ["/registry1", "/registry2"]
    mocked_fetch_definition_mapping.return_value = "__MAPPING__"
    mocked_resolve_context.return_value = wiz_context
    mocked_fetch_package_request_from_command.return_value = "__PACKAGE__"
    mocked_resolve_command.return_value = "__RESOLVED_COMMAND__"

    runner = CliRunner()
    result = runner.invoke(
        wiz.command_line.main, ["run", "fooExe", "--view"] + options
    )
    assert not result.exception
    assert result.exit_code == 0
    assert result.output == (
        "\n"
        "Registries    \n"
        "--------------\n"
        "[0] /registry1\n"
        "[1] /registry2\n"
        "\n"
        "\n"
        "Package          Version   Registry   Description       \n"
        "--------------   -------   --------   ------------------\n"
        "foo              0.1.0     0          This is Foo 0.1.0.\n"
        "bim [Variant1]   0.1.1     1          This is Bim 0.1.1.\n"
        "\n"
        "\n"
        "Command       Value      \n"
        "-----------   -----------\n"
        "fooExe        foo        \n"
        "fooExeDebug   foo --debug\n"
        "\n"
        "\n"
        "Environment Variable   Environment Value\n"
        "--------------------   -----------------\n"
        "KEY1                   value1           \n"
        "KEY2                   value2           \n"
        "\n"
    )

    mocked_fetch_definition_mapping.assert_called_once_with(
        ["/registry1", "/registry2"],
        system_mapping="__SYSTEM__", max_depth=None
    )

    mocked_fetch_package_request_from_command.assert_called_once_with(
        "fooExe", "__MAPPING__"
    )

    mocked_resolve_context.assert_called_once_with(
        ["__PACKAGE__"], "__MAPPING__",
        ignore_implicit=False,
        environ_mapping={},
        maximum_combinations=max_combinations,
        maximum_attempts=max_attempts,
    )

    mocked_resolve_command.assert_not_called()
    mocked_spawn_execute.assert_not_called()
    mocked_history_record_action.assert_not_called()
    logger.error.assert_not_called()


@pytest.mark.parametrize("options, max_combinations, max_attempts", [
    ([], 10, 15),
    (["-mc", "1"], 1, 15),
    (["-ma", "1"], 10, 1),
], ids=[
    "simple",
    "with-maximum-combinations",
    "with-maximum-attempts",
])
def test_run_view_empty(
    mocked_system_query, mocked_registry_fetch, mocked_fetch_definition_mapping,
    mocked_fetch_package_request_from_command, mocked_resolve_context,
    mocked_resolve_command, mocked_spawn_execute, mocked_history_record_action,
    logger, options, max_combinations, max_attempts
):
    """View an empty resolved context from a command execution."""
    mocked_system_query.return_value = "__SYSTEM__"
    mocked_registry_fetch.return_value = ["/registry1", "/registry2"]
    mocked_fetch_definition_mapping.return_value = "__MAPPING__"
    mocked_resolve_context.return_value = {
        "command": {},
        "environ": {},
        "packages": [],
        "registries": ["/registry1", "/registry2"]
    }
    mocked_fetch_package_request_from_command.return_value = "__PACKAGE__"
    mocked_resolve_command.return_value = "__RESOLVED_COMMAND__"

    runner = CliRunner()
    result = runner.invoke(
        wiz.command_line.main, ["run", "fooExe", "--view"] + options
    )
    assert result.exit_code == 0
    assert not result.exception
    assert result.output == (
        "\n"
        "Registries    \n"
        "--------------\n"
        "[0] /registry1\n"
        "[1] /registry2\n"
        "\n"
        "\n"
        "Package   Version   Registry   Description\n"
        "-------   -------   --------   -----------\n"
        "No packages to display.\n"
        "\n"
        "\n"
        "Command   Value\n"
        "-------   -----\n"
        "No commands to display.\n"
        "\n"
        "\n"
        "Environment Variable   Environment Value\n"
        "--------------------   -----------------\n"
        "No environment variables to display.\n"
        "\n"
    )

    mocked_fetch_definition_mapping.assert_called_once_with(
        ["/registry1", "/registry2"],
        system_mapping="__SYSTEM__", max_depth=None
    )

    mocked_fetch_package_request_from_command.assert_called_once_with(
        "fooExe", "__MAPPING__"
    )

    mocked_resolve_context.assert_called_once_with(
        ["__PACKAGE__"], "__MAPPING__",
        ignore_implicit=False,
        environ_mapping={},
        maximum_combinations=max_combinations,
        maximum_attempts=max_attempts,
    )

    mocked_resolve_command.assert_not_called()
    mocked_spawn_execute.assert_not_called()
    mocked_history_record_action.assert_not_called()
    logger.error.assert_not_called()


@pytest.mark.parametrize("options, max_combinations, max_attempts", [
    ([], 10, 15),
    (["-mc", "1"], 1, 15),
    (["-ma", "1"], 10, 1),
], ids=[
    "simple",
    "with-maximum-combinations",
    "with-maximum-attempts",
])
def test_run_with_resolution_error(
    mocked_system_query, mocked_registry_fetch, mocked_fetch_definition_mapping,
    mocked_fetch_package_request_from_command, mocked_resolve_context,
    mocked_resolve_command, mocked_spawn_execute, mocked_history_record_action,
    logger, options, max_combinations, max_attempts
):
    """Fail to resolve a context."""
    exception = wiz.exception.WizError("Oh Shit!")
    mocked_system_query.return_value = "__SYSTEM__"
    mocked_registry_fetch.return_value = ["/registry1", "/registry2"]
    mocked_fetch_definition_mapping.return_value = "__MAPPING__"
    mocked_fetch_package_request_from_command.return_value = "__PACKAGE__"
    mocked_resolve_context.side_effect = exception

    runner = CliRunner()
    result = runner.invoke(wiz.command_line.main, ["run", "fooExe"] + options)
    assert result.exit_code == 0
    assert not result.exception
    assert result.output == ""

    mocked_fetch_definition_mapping.assert_called_once_with(
        ["/registry1", "/registry2"],
        system_mapping="__SYSTEM__", max_depth=None
    )

    mocked_resolve_context.assert_called_once_with(
        ["__PACKAGE__"], "__MAPPING__",
        ignore_implicit=False,
        environ_mapping={},
        maximum_combinations=max_combinations,
        maximum_attempts=max_attempts,
    )

    mocked_fetch_package_request_from_command.assert_called_once_with(
        "fooExe", "__MAPPING__"
    )

    mocked_resolve_command.assert_not_called()
    mocked_spawn_execute.assert_not_called()

    mocked_history_record_action.assert_called_once_with(
        "RAISE_EXCEPTION", error=exception
    )

    logger.error.assert_called_once_with("Oh Shit!")


@pytest.mark.parametrize("options, max_combinations, max_attempts", [
    ([], 10, 15),
    (["-mc", "1"], 1, 15),
    (["-ma", "1"], 10, 1),
], ids=[
    "simple",
    "with-maximum-combinations",
    "with-maximum-attempts",
])
def test_run_initial_environment(
    mocked_system_query, mocked_registry_fetch, mocked_fetch_definition_mapping,
    mocked_fetch_package_request_from_command, mocked_resolve_context,
    mocked_resolve_command, mocked_spawn_execute, wiz_context, options,
    max_combinations, max_attempts
):
    """Execute a command to extend an initial environment."""
    mocked_system_query.return_value = "__SYSTEM__"
    mocked_registry_fetch.return_value = ["/registry1", "/registry2"]
    mocked_fetch_definition_mapping.return_value = "__MAPPING__"
    mocked_resolve_context.return_value = wiz_context
    mocked_fetch_package_request_from_command.return_value = "__PACKAGE__"
    mocked_resolve_command.return_value = "__RESOLVED_COMMAND__"

    runner = CliRunner()
    result = runner.invoke(
        wiz.command_line.main, [
            "--init", "PATH=/path", "--init", "PYTHONPATH=/other-path",
            "run", "fooExe"
        ] + options
    )
    assert result.exit_code == 0
    assert not result.exception
    assert result.output == ""

    mocked_resolve_context.assert_called_once_with(
        ["__PACKAGE__"], "__MAPPING__", ignore_implicit=False,
        environ_mapping={"PATH": "/path", "PYTHONPATH": "/other-path"},
        maximum_combinations=max_combinations,
        maximum_attempts=max_attempts,
    )

    mocked_spawn_execute.assert_called_once_with(
        "__RESOLVED_COMMAND__", {"KEY1": "value1", "KEY2": "value2"}
    )


@pytest.mark.parametrize("options, recorded", [
    ([], False),
    (["--record", tempfile.gettempdir()], True)
], ids=[
    "normal",
    "recorded",
])
@pytest.mark.usefixtures("mock_datetime_now")
@pytest.mark.usefixtures("mocked_system_query")
@pytest.mark.usefixtures("mocked_registry_fetch")
@pytest.mark.usefixtures("mocked_fetch_definition_mapping")
@pytest.mark.usefixtures("mocked_resolve_context")
@pytest.mark.usefixtures("mocked_export_definition")
@pytest.mark.usefixtures("mocked_export_script")
@pytest.mark.usefixtures("mocked_history_record_action")
def test_freeze_recorded(
    mocked_history_start_recording, mocked_history_get,
    mocked_filesystem_export, mocked_click_prompt, options, recorded
):
    """Record history when freezing a resolved environment."""
    mocked_history_get.return_value = "__HISTORY__"
    mocked_click_prompt.side_effect = ["foo", "This is a description", "0.1.0"]

    runner = CliRunner()
    result = runner.invoke(
        wiz.command_line.main,
        options + ["freeze", "foo", "-o", "/output/path"],
    )
    assert result.exit_code == 0
    assert not result.exception
    assert result.output == ""

    if recorded:
        mocked_history_start_recording.assert_called_once_with(
            command=" ".join(
                ["wiz"] + options + ["freeze", "foo", "-o", "/output/path"]
            )
        )
        mocked_history_get.assert_called_once_with(serialized=True)
        mocked_filesystem_export.assert_called_once_with(
            tempfile.gettempdir() + "/wiz-NOW.dump", "__HISTORY__",
            compressed=True
        )

    else:
        mocked_history_start_recording.assert_not_called()
        mocked_history_get.assert_not_called()
        mocked_filesystem_export.assert_not_called()


@pytest.mark.parametrize("options", [
    [], ["--format", "wiz"]
], ids=[
    "default",
    "with-option",
])
def test_freeze(
    mocked_system_query, mocked_registry_fetch, mocked_fetch_definition_mapping,
    mocked_resolve_context, mocked_export_definition, mocked_export_script,
    mocked_history_record_action, logger, mocked_click_prompt, wiz_context,
    options
):
    """Freeze a resolved environment."""
    mocked_system_query.return_value = "__SYSTEM__"
    mocked_registry_fetch.return_value = ["/registry1", "/registry2"]
    mocked_fetch_definition_mapping.return_value = "__MAPPING__"
    mocked_resolve_context.return_value = wiz_context
    mocked_click_prompt.side_effect = ["foo", "This is a description.", "0.1.0"]

    runner = CliRunner()
    result = runner.invoke(
        wiz.command_line.main,
        ["freeze", "foo", "-o", "/output/path"] + options,
    )
    assert result.exit_code == 0
    assert not result.exception
    assert result.output == ""

    mocked_fetch_definition_mapping.assert_called_once_with(
        ["/registry1", "/registry2"],
        system_mapping="__SYSTEM__", max_depth=None
    )

    mocked_resolve_context.assert_called_once_with(
        ["foo"], "__MAPPING__", ignore_implicit=False, environ_mapping={},
    )

    mocked_export_definition.assert_called_once_with(
        "/output/path",
        {
            "identifier": "foo",
            "version": "0.1.0",
            "description": "This is a description.",
            "command": {
                "fooExe": "foo",
                "fooExeDebug": "foo --debug"
            },
            "environ": {
                "KEY1": "value1",
                "KEY2": "value2",
            }
        }
    )

    mocked_export_script.assert_not_called()
    mocked_history_record_action.assert_not_called()
    logger.error.assert_not_called()
    logger.warning.assert_not_called()


@pytest.mark.parametrize("options", [
    [], ["--format", "wiz"]
], ids=[
    "default",
    "with-option",
])
def test_freeze_empty(
    mocked_system_query, mocked_registry_fetch, mocked_fetch_definition_mapping,
    mocked_resolve_context, mocked_export_definition, mocked_export_script,
    mocked_history_record_action, logger, mocked_click_prompt, options
):
    """Freeze an empty resolved environment."""
    mocked_system_query.return_value = "__SYSTEM__"
    mocked_registry_fetch.return_value = ["/registry1", "/registry2"]
    mocked_fetch_definition_mapping.return_value = "__MAPPING__"
    mocked_resolve_context.return_value = {}
    mocked_click_prompt.side_effect = ["foo", "This is a description.", "0.1.0"]

    runner = CliRunner()
    result = runner.invoke(
        wiz.command_line.main,
        ["freeze", "foo", "-o", "/output/path"] + options,
    )
    assert result.exit_code == 0
    assert not result.exception
    assert result.output == ""

    mocked_fetch_definition_mapping.assert_called_once_with(
        ["/registry1", "/registry2"],
        system_mapping="__SYSTEM__", max_depth=None
    )

    mocked_resolve_context.assert_called_once_with(
        ["foo"], "__MAPPING__", ignore_implicit=False, environ_mapping={},
    )

    mocked_export_definition.assert_called_once_with(
        "/output/path",
        {
            "identifier": "foo",
            "version": "0.1.0",
            "description": "This is a description."
        }
    )

    mocked_export_script.assert_not_called()
    mocked_history_record_action.assert_not_called()
    logger.error.assert_not_called()
    logger.warning.assert_not_called()


@pytest.mark.parametrize("format_name, options, command", [
    ("tcsh", ["--format", "tcsh"], None),
    ("tcsh", ["--format", "tcsh"], "fooExe"),
    ("bash", ["--format", "bash"], None),
    ("bash", ["--format", "bash"], "fooExe"),
], ids=[
    "tcsh",
    "tcsh-with-command",
    "bash",
    "bash-with-command",
])
def test_freeze_as_script(
    mocked_system_query, mocked_registry_fetch, mocked_fetch_definition_mapping,
    mocked_resolve_context, mocked_export_definition, mocked_export_script,
    mocked_history_record_action, logger, mocked_click_prompt, wiz_context,
    format_name, options, command
):
    """Freeze a resolved environment as a script."""
    mocked_system_query.return_value = "__SYSTEM__"
    mocked_registry_fetch.return_value = ["/registry1", "/registry2"]
    mocked_fetch_definition_mapping.return_value = "__MAPPING__"
    mocked_resolve_context.return_value = wiz_context
    mocked_click_prompt.side_effect = ["foo", command]

    runner = CliRunner()
    result = runner.invoke(
        wiz.command_line.main,
        ["freeze", "foo", "-o", "/output/path"] + options,
    )
    assert result.exit_code == 0
    assert not result.exception
    assert result.output == (
        "Available aliases:\n"
        "- foo\n"
        "- foo --debug\n"
    )

    mocked_fetch_definition_mapping.assert_called_once_with(
        ["/registry1", "/registry2"],
        system_mapping="__SYSTEM__", max_depth=None
    )

    mocked_resolve_context.assert_called_once_with(
        ["foo"], "__MAPPING__", ignore_implicit=False, environ_mapping={},
    )

    mocked_export_script.assert_called_once_with(
        "/output/path", format_name, "foo",
        environ={"KEY1": "value1", "KEY2": "value2"},
        command=command,
        packages=wiz_context["packages"]
    )

    mocked_export_definition.assert_not_called()
    mocked_history_record_action.assert_not_called()
    logger.error.assert_not_called()
    logger.warning.assert_not_called()


@pytest.mark.parametrize("format_name, options, command", [
    ("tcsh", ["--format", "tcsh"], None),
    ("tcsh", ["--format", "tcsh"], "fooExe"),
    ("bash", ["--format", "bash"], None),
    ("bash", ["--format", "bash"], "fooExe"),
], ids=[
    "tcsh",
    "tcsh-with-command",
    "bash",
    "bash-with-command",
])
def test_freeze_as_script_without_commands(
    mocked_system_query, mocked_registry_fetch, mocked_fetch_definition_mapping,
    mocked_resolve_context, mocked_export_definition, mocked_export_script,
    mocked_history_record_action, logger, mocked_click_prompt, wiz_context,
    format_name, options, command
):
    """Freeze a resolved environment as a script."""
    wiz_context["command"] = {}

    mocked_system_query.return_value = "__SYSTEM__"
    mocked_registry_fetch.return_value = ["/registry1", "/registry2"]
    mocked_fetch_definition_mapping.return_value = "__MAPPING__"
    mocked_resolve_context.return_value = wiz_context
    mocked_click_prompt.side_effect = ["foo", command]

    runner = CliRunner()
    result = runner.invoke(
        wiz.command_line.main,
        ["freeze", "foo", "-o", "/output/path"] + options,
    )
    assert result.exit_code == 0
    assert not result.exception
    assert result.output == ""

    mocked_fetch_definition_mapping.assert_called_once_with(
        ["/registry1", "/registry2"],
        system_mapping="__SYSTEM__", max_depth=None
    )

    mocked_resolve_context.assert_called_once_with(
        ["foo"], "__MAPPING__", ignore_implicit=False, environ_mapping={},
    )

    mocked_export_script.assert_called_once_with(
        "/output/path", format_name, "foo",
        environ={"KEY1": "value1", "KEY2": "value2"},
        command=command,
        packages=wiz_context["packages"]
    )

    mocked_export_definition.assert_not_called()
    mocked_history_record_action.assert_not_called()
    logger.error.assert_not_called()
    logger.warning.assert_not_called()


def test_freeze_with_resolution_error(
    mocked_system_query, mocked_registry_fetch, mocked_fetch_definition_mapping,
    mocked_resolve_context, mocked_export_definition, mocked_export_script,
    mocked_history_record_action, logger, mocked_click_prompt
):
    """Fail to resolve a context."""
    exception = wiz.exception.WizError("Oh Shit!")
    mocked_system_query.return_value = "__SYSTEM__"
    mocked_registry_fetch.return_value = ["/registry1", "/registry2"]
    mocked_fetch_definition_mapping.return_value = "__MAPPING__"
    mocked_fetch_package_request_from_command.return_value = "__PACKAGE__"
    mocked_resolve_context.side_effect = exception

    runner = CliRunner()
    result = runner.invoke(
        wiz.command_line.main, ["freeze", "foo", "-o", "/output/path"],
    )
    assert result.exit_code == 0
    assert not result.exception
    assert result.output == ""

    mocked_fetch_definition_mapping.assert_called_once_with(
        ["/registry1", "/registry2"],
        system_mapping="__SYSTEM__", max_depth=None
    )

    mocked_resolve_context.assert_called_once_with(
        ["foo"], "__MAPPING__", ignore_implicit=False, environ_mapping={},
    )

    mocked_click_prompt.assert_not_called()
    mocked_export_definition.assert_not_called()
    mocked_export_script.assert_not_called()

    mocked_history_record_action.assert_called_once_with(
        "RAISE_EXCEPTION", error=exception
    )

    logger.error.assert_called_once_with("Oh Shit!")
    logger.warning.assert_not_called()


@pytest.mark.parametrize("options", [
    ["--", "--incorrect"],
    ["--incorrect"],
], ids=[
    "extra-arguments",
    "unknown-arguments",
])
@pytest.mark.usefixtures("mocked_system_query")
@pytest.mark.usefixtures("mocked_registry_fetch")
@pytest.mark.usefixtures("mocked_fetch_definition_mapping")
@pytest.mark.usefixtures("mocked_resolve_context")
@pytest.mark.usefixtures("mocked_export_definition")
@pytest.mark.usefixtures("mocked_export_script")
@pytest.mark.usefixtures("mocked_history_record_action")
def test_freeze_error(options):
    """Fail to freeze resolved environment."""
    runner = CliRunner()
    result = runner.invoke(wiz.command_line.main, ["freeze", "foo"] + options)
    assert result.exit_code == 2
    assert result.exception


@pytest.mark.usefixtures("mocked_export_definition")
def test_freeze_initial_environment(
    mocked_system_query, mocked_registry_fetch, mocked_fetch_definition_mapping,
    mocked_resolve_context, wiz_context, mocked_click_prompt
):
    """Freeze a resolved environment with initial environment."""
    mocked_system_query.return_value = "__SYSTEM__"
    mocked_registry_fetch.return_value = ["/registry1", "/registry2"]
    mocked_fetch_definition_mapping.return_value = "__MAPPING__"
    mocked_resolve_context.return_value = wiz_context
    mocked_click_prompt.side_effect = ["foo", "This is a description.", "0.1.0"]

    runner = CliRunner()
    result = runner.invoke(
        wiz.command_line.main, [
           "--init", "PATH=/path", "--init", "PYTHONPATH=/other-path",
           "freeze", "foo", "-o", "/output/path"
        ],
    )
    assert result.exit_code == 0
    assert not result.exception
    assert result.output == ""

    mocked_resolve_context.assert_called_once_with(
        ["foo"], "__MAPPING__", ignore_implicit=False,
        environ_mapping={"PATH": "/path", "PYTHONPATH": "/other-path"},
    )


@pytest.mark.parametrize("options, recorded", [
    ([], False),
    (["--record", tempfile.gettempdir()], True)
], ids=[
    "normal",
    "recorded",
])
@pytest.mark.usefixtures("mock_datetime_now")
@pytest.mark.usefixtures("mocked_system_query")
@pytest.mark.usefixtures("mocked_registry_fetch")
@pytest.mark.usefixtures("mocked_fetch_definition_mapping")
@pytest.mark.usefixtures("mocked_registry_install_to_path")
@pytest.mark.usefixtures("mocked_click_confirm")
@pytest.mark.usefixtures("mocked_history_record_action")
def test_install_recorded(
    mocked_history_start_recording, mocked_history_get,
    mocked_filesystem_export, options, recorded
):
    """Record history when installing definition(s)."""
    mocked_history_get.return_value = "__HISTORY__"

    runner = CliRunner()
    result = runner.invoke(
        wiz.command_line.main,
        options + ["install", "/path/to/foo.json", "-o", "/somewhere"],
    )
    assert result.exit_code == 0
    assert not result.exception
    assert result.output == ""

    if recorded:
        mocked_history_start_recording.assert_called_once_with(
            command=" ".join(
                ["wiz"] + options + [
                    "install", "/path/to/foo.json", "-o", "/somewhere"
                ]
            )
        )
        mocked_history_get.assert_called_once_with(serialized=True)
        mocked_filesystem_export.assert_called_once_with(
            tempfile.gettempdir() + "/wiz-NOW.dump", "__HISTORY__",
            compressed=True
        )

    else:
        mocked_history_start_recording.assert_not_called()
        mocked_history_get.assert_not_called()
        mocked_filesystem_export.assert_not_called()


@pytest.mark.parametrize("options, definitions", [
    (["/foo.json"], ("/foo.json",)),
    (["/foo.json", "/bar.json"], ("/foo.json", "/bar.json"))
], ids=[
    "one-definition",
    "several-definitions"
])
def test_install_to_path(
    mocked_system_query, mocked_registry_fetch, mocked_load_definition,
    mocked_registry_install_to_path, mocked_click_confirm,
    mocked_history_record_action, logger, options, definitions
):
    """Install definition in registry using default plugin."""
    mocked_system_query.return_value = "__SYSTEM__"
    mocked_registry_fetch.return_value = ["/registry1", "/registry2"]
    mocked_load_definition.side_effect = definitions

    runner = CliRunner()
    result = runner.invoke(
        wiz.command_line.main,
        ["install"] + options + ["--output", "/registry"],
    )
    assert not result.exception
    assert result.exit_code == 0
    assert result.output == ""

    assert mocked_load_definition.call_count == len(definitions)
    for definition in definitions:
        mocked_load_definition.assert_any_call(definition)

    mocked_registry_install_to_path.assert_called_once_with(
        list(definitions), "/registry", overwrite=False
    )

    mocked_click_confirm.assert_not_called()
    mocked_history_record_action.assert_not_called()
    logger.warning.assert_not_called()
    logger.error.assert_not_called()


def test_install_overwrite_existing(
    mocked_system_query, mocked_registry_fetch, mocked_load_definition,
    mocked_registry_install_to_path, mocked_click_confirm,
    mocked_history_record_action, logger
):
    """Overwrite definition in registry."""
    mocked_system_query.return_value = "__SYSTEM__"
    mocked_registry_fetch.return_value = ["/registry1", "/registry2"]
    mocked_registry_install_to_path.side_effect = (
        wiz.exception.DefinitionsExist(["'foo' [0.1.0]"]),
        None
    )
    mocked_click_confirm.return_value = True

    runner = CliRunner()
    result = runner.invoke(
        wiz.command_line.main,
        ["install", "/foo.json", "--output", "/registry"],
    )
    assert result.exit_code == 0
    assert not result.exception
    assert result.output == ""

    assert mocked_load_definition.call_count == 2
    mocked_load_definition.assert_any_call("/foo.json")

    assert mocked_registry_install_to_path.call_count == 2
    mocked_registry_install_to_path.assert_any_call(
        [mocked_load_definition.return_value], "/registry",
        overwrite=False
    )
    mocked_registry_install_to_path.assert_any_call(
        [mocked_load_definition.return_value], "/registry",
        overwrite=True
    )

    mocked_click_confirm.assert_called_once_with(
        "1 definition(s) already exist in registry.\n"
        "- 'foo' [0.1.0]\n"
        "Overwrite?"
    )

    mocked_history_record_action.assert_not_called()
    logger.warning.assert_not_called()
    logger.error.assert_not_called()


def test_install_skip_existing(
    mocked_system_query, mocked_registry_fetch, mocked_load_definition,
    mocked_registry_install_to_path, mocked_click_confirm,
    mocked_history_record_action, logger
):
    """Skip definition in registry."""
    mocked_system_query.return_value = "__SYSTEM__"
    mocked_registry_fetch.return_value = ["/registry1", "/registry2"]
    mocked_registry_install_to_path.side_effect = (
        wiz.exception.DefinitionsExist(["'foo' [0.1.0]"]),
    )
    mocked_click_confirm.return_value = False

    runner = CliRunner()
    result = runner.invoke(
        wiz.command_line.main,
        ["install", "/foo.json", "--output", "/registry"],
    )
    assert result.exit_code == 0
    assert not result.exception
    assert result.output == ""

    mocked_load_definition.assert_called_once_with("/foo.json")

    mocked_registry_install_to_path.assert_called_once_with(
        [mocked_load_definition.return_value], "/registry",
        overwrite=False
    )

    mocked_click_confirm.assert_called_once_with(
        "1 definition(s) already exist in registry.\n"
        "- 'foo' [0.1.0]\n"
        "Overwrite?"
    )

    mocked_history_record_action.assert_not_called()
    logger.warning.assert_not_called()
    logger.error.assert_not_called()


def test_install_no_change(
    mocked_system_query, mocked_registry_fetch, mocked_load_definition,
    mocked_registry_install_to_path, mocked_click_confirm,
    mocked_history_record_action, logger
):
    """Installation to registry with no changes."""
    mocked_system_query.return_value = "__SYSTEM__"
    mocked_registry_fetch.return_value = ["/registry1", "/registry2"]
    mocked_registry_install_to_path.side_effect = (
        wiz.exception.InstallNoChanges()
    )

    runner = CliRunner()
    result = runner.invoke(
        wiz.command_line.main,
        ["install", "/foo.json", "--output", "/registry"],
    )
    assert result.exit_code == 0
    assert not result.exception
    assert result.output == ""

    mocked_load_definition.assert_called_once_with("/foo.json")

    mocked_registry_install_to_path.assert_called_once_with(
        [mocked_load_definition.return_value], "/registry",
        overwrite=False
    )

    logger.warning.assert_called_once_with("No changes detected in release.")

    mocked_click_confirm.assert_not_called()
    mocked_history_record_action.assert_not_called()
    logger.error.assert_not_called()


def test_install_local_error(
    mocked_system_query, mocked_registry_fetch, mocked_load_definition,
    mocked_registry_install_to_path, mocked_click_confirm,
    mocked_history_record_action, logger
):
    """Fail to install definition to local registry."""
    exception = wiz.exception.WizError("Oh Shit!")
    mocked_system_query.return_value = "__SYSTEM__"
    mocked_registry_fetch.return_value = ["/registry1", "/registry2"]
    mocked_registry_install_to_path.side_effect = exception

    runner = CliRunner()
    result = runner.invoke(
        wiz.command_line.main,
        ["install", "/foo.json", "--output", "/registry"],
    )
    assert result.exit_code == 0
    assert not result.exception
    assert result.output == ""

    mocked_load_definition.assert_called_once_with("/foo.json")

    mocked_registry_install_to_path.assert_called_once_with(
        [mocked_load_definition.return_value], "/registry",
        overwrite=False
    )

    logger.error.assert_called_once_with(exception)

    mocked_history_record_action.assert_called_once_with(
        "RAISE_EXCEPTION", error=exception
    )

    mocked_click_confirm.assert_not_called()
    logger.warning.assert_not_called()


@pytest.mark.parametrize("options", [
    ["--", "--incorrect"],
    ["--incorrect"],
], ids=[
    "extra-arguments",
    "unknown-arguments",
])
@pytest.mark.usefixtures("mocked_system_query")
@pytest.mark.usefixtures("mocked_registry_fetch")
@pytest.mark.usefixtures("mocked_fetch_definition_mapping")
@pytest.mark.usefixtures("mocked_load_definition")
@pytest.mark.usefixtures("mocked_registry_install_to_path")
@pytest.mark.usefixtures("mocked_click_confirm")
@pytest.mark.usefixtures("mocked_history_record_action")
def test_install_local_command_error(options):
    """Fail to install definition to local registry."""
    runner = CliRunner()
    result = runner.invoke(
        wiz.command_line.main,
        ["install"] + options + ["/foo.json", "--output", "/registry"],
    )
    assert result.exit_code == 2
    assert result.exception


@pytest.mark.parametrize("options, recorded", [
    ([], False),
    (["--record", tempfile.gettempdir()], True)
], ids=[
    "normal",
    "recorded",
])
@pytest.mark.usefixtures("mock_datetime_now")
@pytest.mark.usefixtures("mocked_system_query")
@pytest.mark.usefixtures("mocked_registry_fetch")
@pytest.mark.usefixtures("mocked_load_definition")
@pytest.mark.usefixtures("mocked_click_confirm")
@pytest.mark.usefixtures("mocked_click_edit")
@pytest.mark.usefixtures("mocked_history_record_action")
def test_edit_recorded(
    mocked_history_start_recording, mocked_history_get,
    mocked_filesystem_export, options, recorded
):
    """Record history when editing definition(s)."""
    mocked_history_get.return_value = "__HISTORY__"

    runner = CliRunner()
    result = runner.invoke(
        wiz.command_line.main,
        options + ["edit", "/path/to/foo.json"],
    )
    assert result.exit_code == 0
    assert not result.exception
    assert result.output == ""

    if recorded:
        mocked_history_start_recording.assert_called_once_with(
            command=" ".join(
                ["wiz"] + options + ["edit", "/path/to/foo.json"]
            )
        )
        mocked_history_get.assert_called_once_with(serialized=True)
        mocked_filesystem_export.assert_called_once_with(
            tempfile.gettempdir() + "/wiz-NOW.dump", "__HISTORY__",
            compressed=True
        )

    else:
        mocked_history_start_recording.assert_not_called()
        mocked_history_get.assert_not_called()
        mocked_filesystem_export.assert_not_called()


@pytest.mark.usefixtures("mocked_system_query")
@pytest.mark.usefixtures("mocked_registry_fetch")
def test_edit(
    mocked_load_definition, mocked_click_confirm, mocked_click_edit,
    mocked_history_record_action, mocked_filesystem_export, logger
):
    """Edit definition with editor."""
    definition = wiz.definition.Definition(
        {"identifier": "foo"},
        registry_path="/path",
        path="/path/to/foo.json"
    )

    mocked_load_definition.return_value = definition
    mocked_click_edit.return_value = (
        "{\"identifier\": \"foo\",\"version\": \"0.1.0\"}"
    )
    mocked_click_confirm.return_value = True
    mocked_filesystem_export.side_effect = [
        wiz.exception.FileExists("/path"), None
    ]

    runner = CliRunner()
    result = runner.invoke(wiz.command_line.main, ["edit", "/path/to/foo.json"])
    assert result.exit_code == 0
    assert not result.exception
    assert result.output == ""

    mocked_load_definition.assert_called_once_with("/path/to/foo.json")
    mocked_click_edit.assert_called_once_with(
        "{\n"
        "    \"identifier\": \"foo\"\n"
        "}",
        extension=".json"
    )

    mocked_click_confirm.assert_called_once_with("Overwrite 'foo'?")

    assert mocked_filesystem_export.call_count == 2
    mocked_filesystem_export.assert_any_call(
        "/path/to/foo.json",
        (
            "{\n"
            "    \"identifier\": \"foo\",\n"
            "    \"version\": \"0.1.0\"\n"
            "}"
        ),
        overwrite=False
    )
    mocked_filesystem_export.assert_any_call(
        "/path/to/foo.json",
        (
            "{\n"
            "    \"identifier\": \"foo\",\n"
            "    \"version\": \"0.1.0\"\n"
            "}"
        ),
        overwrite=True
    )

    mocked_history_record_action.assert_not_called()

    logger.error.assert_not_called()
    logger.warning.assert_not_called()

    assert logger.info.call_count == 2
    logger.info.assert_any_call("Edit 'foo'.")
    logger.info.assert_any_call("Saved 'foo' in /path/to/foo.json.")


@pytest.mark.usefixtures("mocked_system_query")
@pytest.mark.usefixtures("mocked_registry_fetch")
def test_edit_non_saved(
    mocked_load_definition, mocked_click_confirm, mocked_click_edit,
    mocked_history_record_action, mocked_filesystem_export, logger
):
    """Skip definition editing when updated data is unsaved in editor."""
    definition = wiz.definition.Definition(
        {"identifier": "foo"},
        registry_path="/path",
        path="/path/to/foo.json"
    )

    mocked_load_definition.return_value = definition
    mocked_click_edit.return_value = None

    runner = CliRunner()
    result = runner.invoke(wiz.command_line.main, ["edit", "/path/to/foo.json"])
    assert result.exit_code == 0
    assert not result.exception
    assert result.output == ""

    mocked_load_definition.assert_called_once_with("/path/to/foo.json")
    mocked_click_edit.assert_called_once_with(
        "{\n"
        "    \"identifier\": \"foo\"\n"
        "}",
        extension=".json"
    )

    mocked_click_confirm.assert_not_called()
    mocked_filesystem_export.assert_not_called()
    mocked_history_record_action.assert_not_called()
    logger.error.assert_not_called()

    logger.info.assert_called_once_with("Edit 'foo'.")
    logger.warning.assert_called_once_with("Skip edition for 'foo'.")


@pytest.mark.usefixtures("mocked_system_query")
@pytest.mark.usefixtures("mocked_registry_fetch")
def test_edit_overwrite_existing(
    mocked_load_definition, mocked_click_confirm, mocked_click_edit,
    mocked_history_record_action, mocked_filesystem_export, logger
):
    """Edit definition with editor by overwriting original."""
    definition = wiz.definition.Definition(
        {"identifier": "foo"},
        registry_path="/path",
        path="/path/to/foo.json"
    )

    mocked_load_definition.return_value = definition
    mocked_click_edit.return_value = (
        "{\"identifier\": \"foo\", \"version\": \"0.1.0\"}"
    )

    runner = CliRunner()
    result = runner.invoke(
        wiz.command_line.main,
        ["edit", "/path/to/foo.json", "--overwrite"]
    )
    assert result.exit_code == 0
    assert not result.exception
    assert result.output == ""

    mocked_load_definition.assert_called_once_with("/path/to/foo.json")
    mocked_click_edit.assert_called_once_with(
        "{\n"
        "    \"identifier\": \"foo\"\n"
        "}",
        extension=".json"
    )

    mocked_filesystem_export.assert_called_once_with(
        "/path/to/foo.json",
        (
            "{\n"
            "    \"identifier\": \"foo\",\n"
            "    \"version\": \"0.1.0\"\n"
            "}"
        ),
        overwrite=True
    )

    mocked_click_confirm.assert_not_called()
    mocked_history_record_action.assert_not_called()
    logger.error.assert_not_called()
    logger.warning.assert_not_called()

    assert logger.info.call_count == 2
    logger.info.assert_any_call("Edit 'foo'.")
    logger.info.assert_any_call("Saved 'foo' in /path/to/foo.json.")


@pytest.mark.usefixtures("mocked_system_query")
@pytest.mark.usefixtures("mocked_registry_fetch")
def test_edit_skip_existing(
    mocked_load_definition, mocked_click_confirm, mocked_click_edit,
    mocked_history_record_action, mocked_filesystem_export, logger
):
    """Abort definition edition by skipping overwriting."""
    definition = wiz.definition.Definition(
        {"identifier": "foo"},
        registry_path="/path",
        path="/path/to/foo.json"
    )

    mocked_load_definition.return_value = definition
    mocked_click_edit.return_value = (
        "{\"identifier\": \"foo\", \"version\": \"0.1.0\"}"
    )

    mocked_click_confirm.return_value = False
    mocked_filesystem_export.side_effect = wiz.exception.FileExists("/path")

    runner = CliRunner()
    result = runner.invoke(
        wiz.command_line.main, ["edit", "/path/to/foo.json"]
    )
    assert result.exit_code == 0
    assert not result.exception
    assert result.output == ""

    mocked_load_definition.assert_called_once_with("/path/to/foo.json")
    mocked_click_edit.assert_called_once_with(
        "{\n"
        "    \"identifier\": \"foo\"\n"
        "}",
        extension=".json"
    )

    mocked_filesystem_export.assert_called_once_with(
        "/path/to/foo.json",
        (
            "{\n"
            "    \"identifier\": \"foo\",\n"
            "    \"version\": \"0.1.0\"\n"
            "}"
        ),
        overwrite=False
    )

    mocked_click_confirm.assert_called_once_with("Overwrite 'foo'?")

    mocked_history_record_action.assert_not_called()
    logger.error.assert_not_called()

    logger.warning.assert_called_once_with("Skip edition for 'foo'.")
    logger.info.assert_called_once_with("Edit 'foo'.")


@pytest.mark.usefixtures("mocked_system_query")
@pytest.mark.usefixtures("mocked_registry_fetch")
def test_edit_with_output(
    mocked_load_definition, mocked_click_confirm, mocked_click_edit,
    mocked_history_record_action, mocked_filesystem_export, logger
):
    """Edit definition with editor and save in different output path."""
    definition = wiz.definition.Definition(
        {"identifier": "foo"},
        registry_path="/path",
        path="/path/to/foo.json"
    )

    mocked_load_definition.return_value = definition
    mocked_click_edit.return_value = (
        "{\"identifier\": \"foo\",\"version\": \"0.1.0\"}"
    )

    runner = CliRunner()
    result = runner.invoke(
        wiz.command_line.main,
        ["edit", "/path/to/foo.json", "--output", "/path/to/target"]
    )
    assert result.exit_code == 0
    assert not result.exception
    assert result.output == ""

    mocked_load_definition.assert_called_once_with("/path/to/foo.json")
    mocked_click_edit.assert_called_once_with(
        "{\n"
        "    \"identifier\": \"foo\"\n"
        "}",
        extension=".json"
    )

    mocked_filesystem_export.assert_called_once_with(
        "/path/to/target/foo.json",
        (
            "{\n"
            "    \"identifier\": \"foo\",\n"
            "    \"version\": \"0.1.0\"\n"
            "}"
        ),
        overwrite=False
    )

    mocked_click_confirm.assert_not_called()
    mocked_history_record_action.assert_not_called()
    logger.error.assert_not_called()
    logger.warning.assert_not_called()

    assert logger.info.call_count == 2
    logger.info.assert_any_call("Edit 'foo'.")
    logger.info.assert_any_call("Saved 'foo' in /path/to/target/foo.json.")


@pytest.mark.usefixtures("mocked_system_query")
@pytest.mark.usefixtures("mocked_registry_fetch")
def test_edit_with_operation_set(
    mocked_load_definition, mocked_click_confirm, mocked_click_edit,
    mocked_history_record_action, mocked_filesystem_export, logger
):
    """Edit definition with operation 'set'."""
    definition = wiz.definition.Definition(
        {"identifier": "foo"},
        registry_path="/path",
        path="/path/to/foo.json"
    )

    mocked_load_definition.return_value = definition

    runner = CliRunner()
    result = runner.invoke(
        wiz.command_line.main,
        [
            "edit", "/path/to/foo.json",
            "--set", "install-location", "/path/to/data"
        ]
    )
    assert result.exit_code == 0
    assert not result.exception
    assert result.output == ""

    mocked_load_definition.assert_called_once_with("/path/to/foo.json")

    mocked_filesystem_export.assert_called_once_with(
        "/path/to/foo.json",
        (
            "{\n"
            "    \"identifier\": \"foo\",\n"
            "    \"install-location\": \"/path/to/data\"\n"
            "}"
        ),
        overwrite=False
    )

    mocked_click_confirm.assert_not_called()
    mocked_click_edit.assert_not_called()
    mocked_history_record_action.assert_not_called()
    logger.error.assert_not_called()
    logger.warning.assert_not_called()

    assert logger.info.call_count == 2
    logger.info.assert_any_call("Edit 'foo'.")
    logger.info.assert_any_call("Saved 'foo' in /path/to/foo.json.")


@pytest.mark.usefixtures("mocked_system_query")
@pytest.mark.usefixtures("mocked_registry_fetch")
def test_edit_with_operation_update(
    mocked_load_definition, mocked_click_confirm, mocked_click_edit,
    mocked_history_record_action, mocked_filesystem_export, logger
):
    """Edit definition with operation 'update'."""
    definition = wiz.definition.Definition(
        {
            "identifier": "foo",
            "environ": {
                "KEY1": "VALUE1"
            },
        },
        path="/path/to/foo.json",
        registry_path="/path"
    )

    mocked_load_definition.return_value = definition

    runner = CliRunner()
    result = runner.invoke(
        wiz.command_line.main,
        [
            "edit", "/path/to/foo.json",
            "--update", "environ", "{\"KEY1\": \"VALUE2\"}"
        ]
    )
    assert result.exit_code == 0
    assert not result.exception
    assert result.output == ""

    mocked_load_definition.assert_called_once_with("/path/to/foo.json")

    mocked_filesystem_export.assert_called_once_with(
        "/path/to/foo.json",
        (
            "{\n"
            "    \"identifier\": \"foo\",\n"
            "    \"environ\": {\n"
            "        \"KEY1\": \"VALUE2\"\n"
            "    }\n"
            "}"
        ),
        overwrite=False
    )

    mocked_click_confirm.assert_not_called()
    mocked_click_edit.assert_not_called()
    mocked_history_record_action.assert_not_called()
    logger.error.assert_not_called()
    logger.warning.assert_not_called()

    assert logger.info.call_count == 2
    logger.info.assert_any_call("Edit 'foo'.")
    logger.info.assert_any_call("Saved 'foo' in /path/to/foo.json.")


@pytest.mark.usefixtures("mocked_system_query")
@pytest.mark.usefixtures("mocked_registry_fetch")
def test_edit_with_operation_extend(
    mocked_load_definition, mocked_click_confirm, mocked_click_edit,
    mocked_history_record_action, mocked_filesystem_export, logger
):
    """Edit definition with operation 'extend'."""
    definition = wiz.definition.Definition(
        {
            "identifier": "foo",
            "requirements": [
                "bar >= 0.1.0, < 1"
            ],
        },
        path="/path/to/foo.json",
        registry_path="/path"
    )

    mocked_load_definition.return_value = definition

    runner = CliRunner()
    result = runner.invoke(
        wiz.command_line.main,
        [
            "edit", "/path/to/foo.json",
            "--extend", "requirements", "[\"bim >= 2.5, < 3\", \"baz\"]"
        ]
    )
    assert result.exit_code == 0
    assert not result.exception
    assert result.output == ""

    mocked_load_definition.assert_called_once_with("/path/to/foo.json")

    mocked_filesystem_export.assert_called_once_with(
        "/path/to/foo.json",
        (
            "{\n"
            "    \"identifier\": \"foo\",\n"
            "    \"requirements\": [\n"
            "        \"bar >= 0.1.0, < 1\",\n"
            "        \"bim >= 2.5, < 3\",\n"
            "        \"baz\"\n"
            "    ]\n"
            "}"
        ),
        overwrite=False
    )

    mocked_click_confirm.assert_not_called()
    mocked_click_edit.assert_not_called()
    mocked_history_record_action.assert_not_called()
    logger.error.assert_not_called()
    logger.warning.assert_not_called()

    assert logger.info.call_count == 2
    logger.info.assert_any_call("Edit 'foo'.")
    logger.info.assert_any_call("Saved 'foo' in /path/to/foo.json.")


@pytest.mark.usefixtures("mocked_system_query")
@pytest.mark.usefixtures("mocked_registry_fetch")
def test_edit_with_operation_insert(
    mocked_load_definition, mocked_click_confirm, mocked_click_edit,
    mocked_history_record_action, mocked_filesystem_export, logger
):
    """Edit definition with operation 'insert'."""
    definition = wiz.definition.Definition(
        {
            "identifier": "foo",
            "requirements": [
                "bar >= 0.1.0, < 1",
                "bim >= 2.5, < 3"
            ],
        },
        path="/path/to/foo.json",
        registry_path="/path"
    )

    mocked_load_definition.return_value = definition

    runner = CliRunner()
    result = runner.invoke(
        wiz.command_line.main,
        [
            "edit", "/path/to/foo.json",
            "--insert", "requirements", "baz", "1"
        ]
    )
    assert result.exit_code == 0
    assert not result.exception
    assert result.output == ""

    mocked_load_definition.assert_called_once_with("/path/to/foo.json")

    mocked_filesystem_export.assert_called_once_with(
        "/path/to/foo.json",
        (
            "{\n"
            "    \"identifier\": \"foo\",\n"
            "    \"requirements\": [\n"
            "        \"bar >= 0.1.0, < 1\",\n"
            "        \"baz\",\n"
            "        \"bim >= 2.5, < 3\"\n"
            "    ]\n"
            "}"
        ),
        overwrite=False
    )

    mocked_click_confirm.assert_not_called()
    mocked_click_edit.assert_not_called()
    mocked_history_record_action.assert_not_called()
    logger.error.assert_not_called()
    logger.warning.assert_not_called()

    assert logger.info.call_count == 2
    logger.info.assert_any_call("Edit 'foo'.")
    logger.info.assert_any_call("Saved 'foo' in /path/to/foo.json.")


@pytest.mark.usefixtures("mocked_system_query")
@pytest.mark.usefixtures("mocked_registry_fetch")
def test_edit_with_operation_remove(
    mocked_load_definition, mocked_click_confirm, mocked_click_edit,
    mocked_history_record_action, mocked_filesystem_export, logger
):
    """Edit definition with operation 'remove'."""
    definition = wiz.definition.Definition(
        {
            "identifier": "foo",
            "requirements": [
                "bar >= 0.1.0, < 1",
                "bim >= 2.5, < 3"
            ],
        },
        path="/path/to/foo.json",
        registry_path="/path"
    )

    mocked_load_definition.return_value = definition

    runner = CliRunner()
    result = runner.invoke(
        wiz.command_line.main,
        [
            "edit", "/path/to/foo.json",
            "--remove", "requirements"
        ]
    )
    assert result.exit_code == 0
    assert not result.exception
    assert result.output == ""

    mocked_load_definition.assert_called_once_with("/path/to/foo.json")

    mocked_filesystem_export.assert_called_once_with(
        "/path/to/foo.json",
        (
            "{\n"
            "    \"identifier\": \"foo\"\n"
            "}"
        ),
        overwrite=False
    )

    mocked_click_confirm.assert_not_called()
    mocked_click_edit.assert_not_called()
    mocked_history_record_action.assert_not_called()
    logger.error.assert_not_called()
    logger.warning.assert_not_called()

    assert logger.info.call_count == 2
    logger.info.assert_any_call("Edit 'foo'.")
    logger.info.assert_any_call("Saved 'foo' in /path/to/foo.json.")


@pytest.mark.usefixtures("mocked_system_query")
@pytest.mark.usefixtures("mocked_registry_fetch")
def test_edit_with_operation_remove_key(
    mocked_load_definition, mocked_click_confirm, mocked_click_edit,
    mocked_history_record_action, mocked_filesystem_export, logger
):
    """Edit definition with operation 'remove-key'."""
    definition = wiz.definition.Definition(
        {
            "identifier": "foo",
            "environ": {
                "KEY1": "VALUE1",
                "KEY2": "VALUE2",
            },
        },
        path="/path/to/foo.json",
        registry_path="/path"
    )

    mocked_load_definition.return_value = definition

    runner = CliRunner()
    result = runner.invoke(
        wiz.command_line.main,
        [
            "edit", "/path/to/foo.json",
            "--remove-key", "environ", "KEY2"
        ]
    )
    assert result.exit_code == 0
    assert not result.exception
    assert result.output == ""

    mocked_load_definition.assert_called_once_with("/path/to/foo.json")

    mocked_filesystem_export.assert_called_once_with(
        "/path/to/foo.json",
        (
            "{\n"
            "    \"identifier\": \"foo\",\n"
            "    \"environ\": {\n"
            "        \"KEY1\": \"VALUE1\"\n"
            "    }\n"
            "}"
        ),
        overwrite=False
    )

    mocked_click_confirm.assert_not_called()
    mocked_click_edit.assert_not_called()
    mocked_history_record_action.assert_not_called()
    logger.error.assert_not_called()
    logger.warning.assert_not_called()

    assert logger.info.call_count == 2
    logger.info.assert_any_call("Edit 'foo'.")
    logger.info.assert_any_call("Saved 'foo' in /path/to/foo.json.")


@pytest.mark.usefixtures("mocked_system_query")
@pytest.mark.usefixtures("mocked_registry_fetch")
def test_edit_with_operation_remove_index(
    mocked_load_definition, mocked_click_confirm, mocked_click_edit,
    mocked_history_record_action, mocked_filesystem_export, logger
):
    """Edit definition with operation 'remove-index'."""
    definition = wiz.definition.Definition(
        {
            "identifier": "foo",
            "requirements": [
                "bar >= 0.1.0, < 1",
                "bim >= 2.5, < 3"
            ],
        },
        path="/path/to/foo.json",
        registry_path="/path"

    )

    mocked_load_definition.return_value = definition

    runner = CliRunner()
    result = runner.invoke(
        wiz.command_line.main,
        [
            "edit", "/path/to/foo.json",
            "--remove-index", "requirements", "0"
        ]
    )
    assert result.exit_code == 0
    assert not result.exception
    assert result.output == ""

    mocked_load_definition.assert_called_once_with("/path/to/foo.json")

    mocked_filesystem_export.assert_called_once_with(
        "/path/to/foo.json",
        (
            "{\n"
            "    \"identifier\": \"foo\",\n"
            "    \"requirements\": [\n"
            "        \"bim >= 2.5, < 3\"\n"
            "    ]\n"
            "}"
        ),
        overwrite=False
    )

    mocked_click_confirm.assert_not_called()
    mocked_click_edit.assert_not_called()
    mocked_history_record_action.assert_not_called()
    logger.error.assert_not_called()
    logger.warning.assert_not_called()

    assert logger.info.call_count == 2
    logger.info.assert_any_call("Edit 'foo'.")
    logger.info.assert_any_call("Saved 'foo' in /path/to/foo.json.")


@pytest.mark.usefixtures("mocked_system_query")
@pytest.mark.usefixtures("mocked_registry_fetch")
def test_edit_error_raised(
    mocked_load_definition, mocked_click_confirm, mocked_click_edit,
    mocked_history_record_action, mocked_filesystem_export, logger
):
    """Fail to edit definition when error is saved."""
    exception = Exception("Oh Shit!")
    mocked_load_definition.side_effect = exception

    runner = CliRunner()
    result = runner.invoke(wiz.command_line.main, ["edit", "/path/to/foo.json"])
    assert result.exit_code == 0
    assert not result.exception
    assert result.output == ""

    mocked_load_definition.assert_called_once_with("/path/to/foo.json")

    mocked_click_edit.assert_not_called()
    mocked_click_confirm.assert_not_called()
    mocked_filesystem_export.assert_not_called()

    logger.info.assert_not_called()
    logger.warning.assert_not_called()

    logger.error.assert_called_once_with("Oh Shit!")

    mocked_history_record_action.assert_called_once_with(
        "RAISE_EXCEPTION", error=exception
    )



# :coding: utf-8

import platform

import pytest
import distro

import wiz.definition
import wiz.exception
import wiz.system
from wiz.utility import Version


@pytest.fixture()
def mocked_platform_system(mocker):
    """Mocked the platform.system function."""
    return mocker.patch.object(platform, "system")


@pytest.fixture()
def mocked_platform_machine(mocker):
    """Mocked the platform.machine function."""
    return mocker.patch.object(platform, "machine")


@pytest.fixture()
def mocked_platform_linux(mocker):
    """Mocked the distro.linux_distribution function."""
    return mocker.patch.object(distro, "linux_distribution")


@pytest.fixture()
def mocked_platform_mac(mocker):
    """Mocked the platform.mac_ver function."""
    return mocker.patch.object(platform, "mac_ver")


@pytest.fixture()
def mocked_platform_win32(mocker):
    """Mocked the platform.win32_ver function."""
    return mocker.patch.object(platform, "win32_ver")


@pytest.fixture()
def mocked_query_linux(mocker):
    """Mocked the query_linux function."""
    return mocker.patch.object(
        wiz.system, "query_linux", return_value="LINUX_SYSTEM_MAPPING"
    )


@pytest.fixture()
def mocked_query_mac(mocker):
    """Mocked the query_mac function."""
    return mocker.patch.object(
        wiz.system, "query_mac", return_value="MAC_SYSTEM_MAPPING"
    )


@pytest.fixture()
def mocked_query_windows(mocker):
    """Mocked the query_windows function."""
    return mocker.patch.object(
        wiz.system, "query_windows", return_value="WINDOWS_SYSTEM_MAPPING"
    )


@pytest.mark.parametrize("_platform, expected", [
    ("Linux", "LINUX_SYSTEM_MAPPING"),
    ("Darwin", "MAC_SYSTEM_MAPPING"),
    ("Windows", "WINDOWS_SYSTEM_MAPPING")
], ids=[
    "linux",
    "mac",
    "windows",
])
@pytest.mark.usefixtures("mocked_query_linux")
@pytest.mark.usefixtures("mocked_query_mac")
@pytest.mark.usefixtures("mocked_query_windows")
def test_query(mocked_platform_system, _platform, expected):
    """Query system mapping."""
    mocked_platform_system.return_value = _platform
    assert wiz.system.query() == expected


def test_query_platform_error(mocked_platform_system):
    """Fails to query system mapping from unsupported platform."""
    mocked_platform_system.return_value = "incorrect"

    with pytest.raises(wiz.exception.UnsupportedPlatform):
        wiz.system.query()


def test_query_version_error(mocked_platform_system, mocked_platform_linux):
    """Fails to query system mapping from incorrect os version."""
    mocked_platform_system.return_value = "linux"
    mocked_platform_linux.side_effect = wiz.exception.VersionError("Error")

    with pytest.raises(wiz.exception.CurrentSystemError):
        wiz.system.query()


@pytest.mark.parametrize("distribution, architecture, expected", [
    (
        ("centos", "7.3.1611", "Core"), "x86_64",
        {
            "platform": "linux",
            "arch": "x86_64",
            "os": {"name": "centos", "version": Version("7.3.1611")}
        }
    ),
    (
        ("centos", "6.5", "Final"), "x86_64",
        {
            "platform": "linux",
            "arch": "x86_64",
            "os": {"name": "centos", "version": Version("6.5")}
        }
    ),
    (
        ("redhat", "5.7", "Final"), "i386",
        {
            "platform": "linux",
            "arch": "i386",
            "os": {"name": "redhat", "version": Version("5.7")}
        }
    )
], ids=[
    "centos-7-64b",
    "centos-6-64b",
    "redhat-5-32b",
])
def test_query_linux(
    mocked_platform_linux, mocked_platform_machine,
    distribution, architecture, expected
):
    """Query linux system mapping."""
    mocked_platform_linux.return_value = distribution
    mocked_platform_machine.return_value = architecture
    assert wiz.system.query_linux() == expected


@pytest.mark.parametrize("mac_ver, architecture, expected", [
    (
        ("10.13.3", ("", "", ""), ""), "x86_64",
        {
            "platform": "mac",
            "arch": "x86_64",
            "os": {"name": "mac", "version": Version("10.13.3")}
        }
    ),
    (
        ("10.11.6", ("", "", ""), ""), "x86_64",
        {
            "platform": "mac",
            "arch": "x86_64",
            "os": {"name": "mac", "version": Version("10.11.6")}
        }
    )
], ids=[
    "mac-10.13",
    "mac-10.11",
])
def test_query_mac(
    mocked_platform_mac, mocked_platform_machine,
    mac_ver, architecture, expected
):
    """Query mac system mapping."""
    mocked_platform_mac.return_value = mac_ver
    mocked_platform_machine.return_value = architecture
    assert wiz.system.query_mac() == expected


@pytest.mark.parametrize("win32_ver, architecture, expected", [
    (
        ("XP", "5.1.2600", "SP2", "Multiprocessor Free"), "i386",
        {
            "platform": "windows",
            "arch": "i386",
            "os": {"name": "windows", "version": Version("5.1.2600")}
        }
    ),
    (
        ("10", "10.0.16299", "", "Multiprocessor Free"), "AMD64",
        {
            "platform": "windows",
            "arch": "AMD64",
            "os": {"name": "windows", "version": Version("10.0.16299")}
        }
    )
], ids=[
    "windows-xp",
    "windows-10",
])
def test_query_windows(
    mocked_platform_win32, mocked_platform_machine,
    win32_ver, architecture, expected
):
    """Query mac system mapping."""
    mocked_platform_win32.return_value = win32_ver
    mocked_platform_machine.return_value = architecture
    assert wiz.system.query_windows() == expected


@pytest.mark.parametrize("definition_data, expected", [
    ({"identifier": "test"}, True),
    ({"identifier": "test", "system": {"platform": "linux"}}, True),
    ({"identifier": "test", "system": {"platform": "other"}}, False),
    ({"identifier": "test", "system": {"arch": "x86_64"}}, True),
    ({"identifier": "test", "system": {"arch": "i386"}}, False),
    ({"identifier": "test", "system": {"os": "centos"}}, True),
    ({"identifier": "test", "system": {"os": "ubuntu"}}, False),
    ({"identifier": "test", "system": {"os": "el"}}, True),
    ({"identifier": "test", "system": {"os": "centos<7"}}, False),
    ({"identifier": "test", "system": {"os": "centos>=7,<8"}}, True),
], ids=[
    "no-system",
    "filter-linux",
    "filter-other-platform",
    "filter-x86_64",
    "filter-other-arch",
    "filter-centos",
    "filter-ubuntu",
    "filter-el",
    "filter-centos-inf-7",
    "filter-centos-7-to-8",
])
def test_validate_for_centos73_64(definition_data, expected):
    """Validate definition for CentOS 7.3, architecture x86_64."""
    system_mapping = {
        "platform": "linux",
        "arch": "x86_64",
        "os": {
            "name": "centos",
            "version": Version("7.5")
        }
    }

    definition = wiz.definition.Definition(definition_data)
    assert wiz.system.validate(definition, system_mapping) == expected


@pytest.mark.parametrize("definition_data, expected", [
    ({"identifier": "test"}, True),
    ({"identifier": "test", "system": {"platform": "mac"}}, True),
    ({"identifier": "test", "system": {"platform": "other"}}, False),
    ({"identifier": "test", "system": {"arch": "x86_64"}}, True),
    ({"identifier": "test", "system": {"arch": "i386"}}, False),
    ({"identifier": "test", "system": {"os": "mac"}}, True),
    ({"identifier": "test", "system": {"os": "other"}}, False),
    ({"identifier": "test", "system": {"os": "el"}}, False),
    ({"identifier": "test", "system": {"os": "mac<10"}}, False),
    ({"identifier": "test", "system": {"os": "mac>=10,<11"}}, True),
], ids=[
    "no-system",
    "filter-mac",
    "filter-other-platform",
    "filter-x86_64",
    "filter-other-arch",
    "filter-mac-os",
    "filter-other",
    "filter-el",
    "filter-mac-inf-10",
    "filter-mac-10-to-11",
])
def test_validate_for_mac_64(definition_data, expected):
    """Validate definition for MacOS 10.13.3, architecture x86_64."""
    system_mapping = {
        "platform": "mac",
        "arch": "x86_64",
        "os": {
            "name": "mac",
            "version": Version("10.13.3")
        }
    }

    definition = wiz.definition.Definition(definition_data)
    assert wiz.system.validate(definition, system_mapping) == expected


@pytest.mark.parametrize("definition_data, expected", [
    ({"identifier": "test"}, True),
    ({"identifier": "test", "system": {"platform": "windows"}}, True),
    ({"identifier": "test", "system": {"platform": "other"}}, False),
    ({"identifier": "test", "system": {"arch": "AMD64"}}, True),
    ({"identifier": "test", "system": {"arch": "i386"}}, False),
    ({"identifier": "test", "system": {"os": "windows"}}, True),
    ({"identifier": "test", "system": {"os": "other"}}, False),
    ({"identifier": "test", "system": {"os": "el"}}, False),
    ({"identifier": "test", "system": {"os": "windows<10"}}, False),
    ({"identifier": "test", "system": {"os": "windows>=10,<11"}}, True),
], ids=[
    "no-system",
    "filter-windows",
    "filter-other-platform",
    "filter-AMD64",
    "filter-other-arch",
    "filter-windows",
    "filter-other",
    "filter-el",
    "filter-mac-inf-10",
    "filter-mac-10-to-11",
])
def test_validate_for_windows_amd64(definition_data, expected):
    """Validate definition for Windows 10.0.16299, architecture AMD64."""
    system_mapping = {
        "platform": "windows",
        "arch": "AMD64",
        "os": {
            "name": "windows",
            "version": Version("10.0.16299")
        }
    }

    definition = wiz.definition.Definition(definition_data)
    assert wiz.system.validate(definition, system_mapping) == expected


def test_validate_requirement_error():
    """Fails to validate definition when os requirement is incorrect."""
    definition = wiz.definition.Definition({
        "identifier": "test",
        "system": {"os": "!!!"}
    })

    with pytest.raises(wiz.exception.DefinitionError):
        wiz.system.validate(definition, {})
